import zmq
import msgpack
from . import nodeio_context
import psycopg2

__all__ = ["Outputs"]


class Outputs:

    __local_instances = []
    __db_write = None
    __bucket = None

    @classmethod
    def _register(cls, instance):
        cls.__local_instances.append(instance)
        return len(cls.__local_instances) - 1

    @classmethod
    def _schema(cls):
        outputList = []
        for v in cls.__local_instances:
            outputList.append(
                {
                    "description": v.description,
                    "units": v.units,
                    "retention": v.retention,
                }
            )
        return outputList

    @classmethod
    def _connections(cls):
        return cls.__local_instances[0].connections

    def __init__(self):
        self.__index = self._register(self)
        self.description = {}
        self.units = {}
        self.retention = {}
        self.values = {}
        self.connections = {}
        self.db_write_api = None

    def define(self, name, units="", description="", retention=0):
        if name.find(" ") != -1 or name.find("/") != -1:
            raise ValueError(
                "Input name cannot contain spaces or / characters. "
                "Use underscore_names or camelCase."
            )
        self.description[name] = description
        self.units[name] = units
        self.retention[name] = retention
        self.values[name] = []
        # open local communication if part of workflow
        if nodeio_context.playbook is not None:
            my_output = nodeio_context.playbook.get(
                f"#{nodeio_context.entropy_identity}/{name}"
            ).decode()
            context = nodeio_context.zmq_context()
            socket = context.socket(zmq.PUB)
            socket.setsockopt(zmq.LINGER, 0)
            socket.bind(my_output)
            self.connections[name] = socket

    def set(self, **kwargs):
        """
        Simple setting of output, that accepts multiple variables at the same time.

        Depending on the retention specified when output variable was defined
        this value will just be available to other nodes, or it will also be
        saved to the database.
        """
        for key, value in kwargs.items():
            assert key in self.values, (
                f"There is no field name '{key}' in Output fields.\n"
                f"All output field names have to be defined at the start."
            )
            # write internally

            encoded_value = msgpack.packb(value)

            if key in self.connections:
                self.connections[key].send(encoded_value)

            if nodeio_context.runtime_data is not None:
                if self.retention[key] > 0:
                    # if retention is non-zero send to database

                    # TODO: use zeromq bridge to SQL to simplify nodeio logic
                    # and have one place to control reliability
                    nodeio_context.runtime_data.execute(
                        f"""INSERT INTO "#{nodeio_context.entropy_identity}/{key}" """
                        f""" (time, value) VALUES (NOW(), msgpack_decode(%s));
                            """,
                        (psycopg2.Binary(encoded_value)),
                    )
            else:
                # for debugging purposes
                print(
                    f"\tENTROPYLAB - OUTPUT #{nodeio_context.node_name}/{key}"
                    f" SAVE (retention = {self.retention[key]})"
                )
                print(f"\t{value}")
        return

    # TODO: add also write that will have more controllable options when writing data



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["QPUcircuitRunner"]


class QPUcircuitRunner(object):
    def __init__(
        self, workflow_node_unique_name, circuit_param=None, error_correction=None
    ):
        """Executes given circuit sequence

        :param circuit_param: (JSON - STREAM) cirucuit description
        :param error_correction: (JSON - STREAM) corrected data
        """
        self._command = "python3"
        self._bin = "circuitExec.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/cpu.svg"
        self._inputs = _Inputs(
            circuit_param=circuit_param,
            error_correction=error_correction,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, circuit_param=None, error_correction=None):
        self._inputs = Inputs()

        self._inputs.state(
            "circuit_param", description="cirucuit description", units="JSON"
        )
        self._inputs.set(circuit_param=circuit_param)

        self._inputs.state(
            "error_correction", description="corrected data", units="JSON"
        )
        self._inputs.set(error_correction=error_correction)

    @property
    def circuit_param(self):
        """Input: cirucuit description (JSON)"""
        return self._inputs.get("circuit_param")

    @circuit_param.setter
    def circuit_param(self, value):
        """Input: cirucuit description (JSON)"""
        self._inputs.set(circuit_param=value)

    @property
    def error_correction(self):
        """Input: corrected data (JSON)"""
        return self._inputs.get("error_correction")

    @error_correction.setter
    def error_correction(self, value):
        """Input: corrected data (JSON)"""
        self._inputs.set(error_correction=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "precorrected_data",
            "circuit_finished",
            "final_data",
        ]

    @property
    def precorrected_data(self):
        """Output: data send for error correction
        :return: (JSON)
        """
        return "#" + self._name + "/precorrected_data"

    @property
    def circuit_finished(self):
        """Output: Event trigger
        :return: (bool)
        """
        return "#" + self._name + "/circuit_finished"

    @property
    def final_data(self):
        """Output: circuit averaged output
        :return: (bool)
        """
        return "#" + self._name + "/final_data"



import entropylab.flame.nodeio as nodeio

import numpy as np
import matplotlib.pyplot as plt
import io
import base64

# ==================== DEFINE NODE ====================

nodeio.context(
    name="FinalDataAnalysis",
    description="Final fitting and plotting",
    icon="bootstrap/file-bar-graph.svg",
)

input = nodeio.Inputs()
input.stream("data", units="JSON", description="experiment results")

output = nodeio.Outputs()
output.define(
    "estimated_parameter", units="MHz", description="final result", retention=2
)
output.define("final_plot", units="png", description="to-do", retention=2)

nodeio.register()  # enables this to be used as part of workflow


# ==================== DRY-RUN DATA ====================

input.set(data=3.23)

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:
    a = input.get("data")
    print(a)

    # do some custom plot
    x = np.linspace(0, 10, 100)
    y = np.cos(a * x / 10)
    plt.figure()
    plt.plot(x, y, "b-")
    plt.plot(x, np.cos(y) / 3, "r-")
    plt.xlabel("Time (s)")
    plt.ylabel("Amplitude (rel.un)")
    plt.legend(["probe1", "drive1"])

    buffer = io.BytesIO()

    plt.savefig(buffer, format="png")

    buffer.seek(0)
    figure_base64_encoded_png = base64.b64encode(buffer.read()).decode()

    output.set(estimated_parameter=y.tolist())
    output.set(final_plot={"png": figure_base64_encoded_png})


# ================= =================== =================



from dash import callback_context
from dash.dependencies import Input, Output, State

from entropylab.pipeline.api.param_store import ParamStore
from entropylab.dashboard.pages.params.utils import (
    param_store_to_df,
    param_store_to_commits_df,
    data_diff,
)


def register_callbacks(app, param_store: ParamStore):
    @app.callback(
        [
            Output("data-table", "data"),
            Output("commit-table", "data"),
            Output("dirty-badge-collapse", "is_open"),
        ],
        Input("data-table", "data"),
        Input("commit-button", "n_clicks"),
        Input("add-key-button", "n_clicks"),
        Input("checkout-button", "n_clicks"),
        Input("save-temp-button", "n_clicks"),
        Input("load-temp-button", "n_clicks"),
        Input("commit-table", "selected_rows"),
        Input("commit-table", "data"),
        # Input("interval", "n_intervals"),
        State("data-table", "active_cell"),
        State("data-table", "data_previous"),
        State("commit-label", "value"),
        prevent_initial_call=True,
    )
    def button_callback(
        data,
        commit_click,
        add_click,
        checkout_click,
        save_temp_click,
        load_temp_click,
        commits_selected_rows,
        commits_data,
        # interval,
        active_cell,
        data_prev,
        commit_label,
    ):
        prop_id = callback_context.triggered[0]["prop_id"].split(".")[0]

        if prop_id == "commit-button":
            param_store.commit(commit_label)
        elif prop_id == "checkout-button":
            if commits_selected_rows and len(commits_selected_rows) > 0:
                commit = commits_data[commits_selected_rows[0]]
                param_store.checkout(commit["commit_id"])
        elif prop_id == "data-table":
            data_diff(param_store, data, data_prev)
        elif prop_id == "add-key-button":
            param_store[f"newKey{len(data)}"] = None
        elif prop_id == "save-temp-button":
            param_store.save_temp()
        elif prop_id == "load-temp-button":
            param_store.load_temp()

        return (
            param_store_to_df(param_store).to_dict("records"),
            param_store_to_commits_df(param_store).to_dict("records"),
            param_store.is_dirty,
        )



import argparse
import importlib

from entropylab.flame.workflow import Workflow

__all__ = ["_get_workflow", "workflow_json", "workflow_summary"]


def _get_workflow(workflow_path):
    workflow_path = workflow_path.replace(".py", "")
    importlib.import_module(workflow_path)
    workflow = Workflow._main_workflow()
    # now prevent workflow to overwrite parameters on exit
    Workflow._prevent_parameter_file_overwrite()
    return workflow


def workflow_json(workflow_path):
    workflow = _get_workflow(workflow_path)

    return workflow._to_json()


def workflow_summary(workflow_path):
    workflow = _get_workflow(workflow_path)
    return workflow._summary_json()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Utilities for workflow")
    parser.add_argument(
        "function",
        type=str,
        default="workflow_json",
        help=" workflow_json = JSON with Cytoscape graph representation;"
        " workflow_summary = JSON with all indexable data for full-text search about workflow",
    )
    parser.add_argument(
        "-w",
        "--workflow",
        type=str,
        default="workflow.py",
        help="Python file that defines main entropylab.Workflow (default workflow.py)",
    )

    args = parser.parse_args()
    if args.function == "workflow_json":
        print(workflow_json(args.workflow))
    elif args.function == "workflow_summary":
        print(workflow_summary(args.workflow))
    else:
        print("ERROR")



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from datetime import timedelta
from enum import Enum, unique
from typing import Dict, List, Optional, MutableMapping

import pandas as pd


@unique
class MergeStrategy(Enum):
    OURS = 1
    THEIRS = 2


class ParamStore(ABC, MutableMapping):
    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def get_value(self, key: str, commit_id: Optional[str] = None) -> object:
        """
        Returns the value of a param by its key

        :param key: the key identifying the param
        :param commit_id: an optional commit_id. If provided, the value will be
        returned from the specified commit
        """
        pass

    @abstractmethod
    def get_param(self, key: str, commit_id: Optional[str] = None) -> Param:
        """
        Returns a copy of the Param instance of a value stored in ParamStore

        :param key: the key identifying the param
        :param commit_id: an optional commit_id. If provided, the Param will be
        returned from the specified commit
        """
        pass

    @abstractmethod
    def set_param(self, key: str, value: object, expiration: Optional[timedelta]):
        """
        Sets a Param in the ParamStore

        :param key: the key identifying the param
        :param value: the value of the param
        :param key: an optional period of time (measured from the time the param
        is committed) after which the prime expires (i.e. param.has_expired
        returns True)
        """
        pass

    @abstractmethod
    def rename_key(self, key: str, new_key: str):
        pass

    @abstractmethod
    def commit(self, label):
        pass

    @abstractmethod
    def checkout(self, commit_id: str, commit_num: int = None, move_by: int = None):
        pass

    @abstractmethod
    def list_commits(self, label: Optional[str]):
        """
        Returns a list of commits

        :param label: an optional label, if given then only commits that match
        it will be returned
        """
        pass

    @abstractmethod
    def list_values(self, key: str) -> pd.DataFrame:
        """
        Lists all the values of a given key taken from commit history,
        sorted by date ascending

        :param key: the key for which to list values
        :returns: a list of tuples where the values are, in order:
            - the value of the key
            - time of commit
            - commit_id
            - label assigned to commit
        """
        pass

    @abstractmethod
    def merge(
        self,
        theirs: Dict | ParamStore,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.OURS,
    ) -> None:
        pass

    """ Tags """

    @abstractmethod
    def add_tag(self, tag: str, key: str) -> None:
        pass

    @abstractmethod
    def remove_tag(self, tag: str, key: str) -> None:
        pass

    @abstractmethod
    def list_keys_for_tag(self, tag: str) -> List[str]:
        pass

    @abstractmethod
    def list_tags_for_key(self, key: str):
        pass

    """ Temporary State """

    @abstractmethod
    def save_temp(self) -> None:
        pass

    @abstractmethod
    def load_temp(self) -> None:
        pass

    @property
    @abstractmethod
    def is_dirty(self):
        """
        True iff params have been changed since the store has last been
        initialized or checked out"""
        pass


class Param(Dict):
    def __init__(self, value):
        super().__init__()
        self.value: object = value
        self.commit_id: Optional[str] = None
        self.expiration: Optional[timedelta | int] = None

    def __repr__(self):
        return (
            f"<Param(value={self.value}, "
            f"commit_id={self.commit_id}, "
            f"expiration={self.__expiration_repr})> "
        )

    @property
    def __expiration_repr(self):
        if isinstance(self.expiration, int):
            return _ns_to_datetime(self.expiration)
        else:
            return False

    @property
    def has_expired(self):
        """
        Indicates whether the Param value has expired. Returns True iff the Param
        has been committed and the time elapsed since the commit operation has exceeded
        the time recorded in the `expiration` property"""
        if isinstance(self.expiration, int):
            return self.expiration < time.time_ns()
        else:
            return False


def _ns_to_datetime(ns: int) -> pd.datetime:
    """
    Converts a UNIX epoch timestamp in nano-seconds to a human readable string"""
    return pd.to_datetime(ns)



from dash import Dash, html, dcc, callback, Output, Input

from entropylab import SqlAlchemyDB
from entropylab.dashboard.pages import results, params
from entropylab.dashboard.pages.results.dashboard_data import (
    SqlalchemyDashboardDataReader,
)
from entropylab.dashboard.theme import (
    theme_stylesheet,
)
from entropylab.logger import logger
from entropylab.pipeline.api.in_process_param_store import InProcessParamStore
from entropylab.pipeline.results_backend.sqlalchemy.project import (
    project_name,
    project_path,
    param_store_file_path,
)


def build_dashboard_app(proj_path):
    """Initialize the dashboard Dash app to show an Entropy project

    :param proj_path path where the Entropy project to be used resides."""

    """ Creating and setting up the Dash app """

    app = Dash(
        __name__,
        external_stylesheets=[theme_stylesheet],
        update_title=None,
        suppress_callback_exceptions=True,
    )
    app.title = f"Entropy - {project_name(proj_path)} [{project_path(proj_path)}]"

    app.layout = html.Div(
        [dcc.Location(id="url", refresh=False), html.Div(id="page-content")]
    )

    """ Initializing data sources """

    dashboard_data_reader = SqlalchemyDashboardDataReader(SqlAlchemyDB(proj_path))
    # noinspection PyBroadException
    try:
        param_store_file = param_store_file_path(proj_path)
        param_store = InProcessParamStore(param_store_file)
    except BaseException as e:
        logger.exception(
            f"Exception when loading ParamStore file at '{param_store_file}'"
        )
        raise e

    """ Building page layouts """

    results_layout = results.layout.build_layout(proj_path, dashboard_data_reader)
    params_layout = params.layout.build_layout(proj_path, param_store)

    """ Registering callbacks used in pages """

    results.callbacks.register_callbacks(app, dashboard_data_reader)
    params.callbacks.register_callbacks(app, param_store)

    """ Callback to route between page layouts based on URL """

    @callback(Output("page-content", "children"), Input("url", "pathname"))
    def display_page(pathname):
        if pathname == "/":
            return results_layout
        elif pathname == "/params":
            return params_layout
        else:
            return "404"

    return app



from __future__ import annotations

import copy
import hashlib
import json
import os.path
import shutil
import string
import threading
import time
from datetime import datetime, timedelta
from pathlib import Path
from random import SystemRandom
from typing import Optional, Dict, Any, List, Callable, Set

import jsonpickle as jsonpickle
import pandas as pd
from tinydb import TinyDB, Query
from tinydb.storages import MemoryStorage, Storage
from tinydb.table import Document

from entropylab.logger import logger
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.api.param_store import ParamStore, MergeStrategy, Param

INFO_DOC_ID = 1
INFO_TABLE = "info"
TEMP_DOC_ID = 1
TEMP_TABLE = "temp"
VERSION_KEY = "version"


class Metadata:
    def __init__(self, d: Dict = None):
        self.id: str = ""  # commit_id
        self.timestamp: int = time.time_ns()  # nanoseconds since epoch
        self.label: Optional[str] = None
        if d:
            self.__dict__.update(d)

    def __repr__(self) -> str:
        d = self.__dict__.copy()
        d["timestamp"] = _ns_to_datetime(self.timestamp)
        return f"<Metadata({_dict_to_json(d)})>"


class JSONPickleStorage(Storage):
    def __init__(self, filename):
        self.filename = filename

    def read(self):
        if not os.path.isfile(self.filename):
            return None
        with open(self.filename) as handle:
            # noinspection PyBroadException
            try:
                s = handle.read()
                data = jsonpickle.decode(s)
                return data
            except BaseException:
                logger.exception(
                    f"Exception decoding TinyDB JSON file '{self.filename}'"
                )
                return None

    def write(self, data):
        # noinspection PyBroadException
        try:
            with open(self.filename, "w+") as handle:
                s = jsonpickle.encode(data)
                handle.write(s)
        except BaseException:
            logger.exception(f"Exception encoding TinyDB JSON file '{self.filename}'")

    def close(self):
        pass


class InProcessParamStore(ParamStore):
    """Naive implementation of ParamStore based on tinydb

    Important:
    Using this implementation in multiple concurrent processes is not supported.
    """

    def __init__(
        self,
        path: Optional[str] | Optional[Path] = None,
        theirs: Optional[Dict | ParamStore] = None,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.THEIRS,
    ):

        super().__init__()

        self.__lock = threading.RLock()
        self.__base_commit_id: Optional[str] = None  # last commit checked out/committed
        self.__base_doc_id: Optional[int] = None  # tinydb document id of last commit...
        self.__params: Dict[str, Param] = dict()  # where current params are stored
        self.__tags: Dict[str, List[str]] = dict()  # tags that are mapped to keys
        self.__is_dirty: bool = True  # can the store be committed at this time?
        self.__dirty_keys: Set[str] = set()  # updated keys not committed yet

        if path is None:
            self.__is_in_memory_mode = True
            self.__db = TinyDB(storage=MemoryStorage)
        else:
            self.__is_in_memory_mode = False
            if isinstance(path, Path):
                path = str(path)
            if not os.path.isfile(path):
                logger.debug(
                    f"Could not find a ParamStore JSON file at '{path}'. "
                    f"A new, empty file will be created."
                )
            self.__db = TinyDB(path, storage=JSONPickleStorage)
        if theirs is not None:
            self.merge(theirs, merge_strategy)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__db.close()

    """ Properties """

    @property
    def is_dirty(self):
        with self.__lock:
            return self.__is_dirty

    """ MutableMapping """

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set self[key] to value. The key-value pair becomes a "param" and
        can be persisted using `commit()` and retrieved later using
        `checkout()`.

        Note: Keys should not start with a dunder (`__`). Such keys are not
        treated as params and are not persisted when `commit()` is called.
        """

        if key.startswith("__") or key.startswith(f"_{self.__class__.__name__}__"):
            # keys that are private attributes are not params and are treated
            # as regular object attributes
            object.__setattr__(self, key, value)
        else:
            with self.__lock:
                self.__params.__setitem__(key, Param(value))
                self.__is_dirty = True
                self.__dirty_keys.add(key)

    def __getitem__(self, key: str) -> Any:
        with self.__lock:
            return self.__params.__getitem__(key).value

    def __delitem__(self, *args, **kwargs):
        with self.__lock:
            key = args[0]
            self.__params.__delitem__(*args, **kwargs)
            self.__remove_key_from_tags(key)
            self.__is_dirty = True
            self.__dirty_keys.add(key)

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            try:
                return self[key]
            except KeyError:
                raise AttributeError(key)

    def __setattr__(self, key, value):
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            try:
                self[key] = value
            except BaseException:
                raise AttributeError(key)
        else:
            object.__setattr__(self, key, value)

    def __iter__(self):
        with self.__lock:
            values = _extract_param_values(self.__params)
            return values.__iter__()

    def __len__(self):
        with self.__lock:
            return self.__params.__len__()

    def __contains__(self, key):
        with self.__lock:
            return self.__params.__contains__(key)

    def __repr__(self):
        with self.__lock:
            return f"<InProcessParamStore({self.to_dict().__repr__()})>"

    def keys(self):
        with self.__lock:
            return self.__params.keys()

    def to_dict(self) -> Dict:
        with self.__lock:
            return _extract_param_values(self.__params)

    def get_value(self, key: str, commit_id: Optional[str] = None) -> object:
        with self.__lock:
            if commit_id is None:
                return self[key]
            else:
                commit = self.__get_commit(commit_id)
                return commit["params"][key].value

    def get_param(self, key: str, commit_id: Optional[str] = None) -> Param:
        with self.__lock:
            if commit_id is None:
                return copy.deepcopy(self.__params[key])
            else:
                commit = self.__get_commit(commit_id)
                return copy.deepcopy(commit["params"][key])

    def set_param(self, key: str, value: object, expiration: Optional[timedelta]):

        with self.__lock:
            if key in self.__params:
                param = self.get_param(key)
            else:
                param = Param(value)
            param.value = value
            if expiration:
                param.expiration = expiration
            else:
                param.expiration = None
            self.__params.__setitem__(key, param)
            self.__is_dirty = True
            self.__dirty_keys.add(key)

    def __remove_key_from_tags(self, key: str):
        for tag in self.__tags:
            if key in self.__tags[tag]:
                self.__tags[tag].remove(key)

    def rename_key(self, key: str, new_key: str):
        with self.__lock:
            if new_key in self.keys():
                raise KeyError(
                    f"Cannot rename key '{key}' to key that already exists: '{new_key}'"
                )
            self.__rename_key_in_tags(key, new_key)
            value = self.__getitem__(key)
            self.__setitem__(new_key, value)
            self.__delitem__(key)

    def __rename_key_in_tags(self, key, new_key):
        for item in self.__tags.items():
            tag = item[0]
            keys = item[1]
            if key in keys:
                self.__tags[tag].remove(key)
                self.__tags[tag].append(new_key)

    """ Commits """

    def commit(self, label: Optional[str] = None) -> str:
        with self.__lock:
            if not self.__is_dirty:
                return self.__base_commit_id
            commit_id = self.__generate_commit_id()
            commit_timestamp = time.time_ns()  # nanoseconds since epoch
            self.__stamp_dirty_params_with_commit(commit_id, commit_timestamp)
            doc = self.__build_document(commit_id, label)
            doc_id = self.__db.insert(doc)
            self.__base_commit_id = doc["metadata"]["id"]
            self.__base_doc_id = doc_id
            self.__is_dirty = False
            self.__dirty_keys.clear()
            return doc["metadata"]["id"]

    def __stamp_dirty_params_with_commit(self, commit_id: str, commit_timestamp: int):
        for key in self.__dirty_keys:
            if key in self.__params:
                param = self.__params[key]
                param.commit_id = commit_id
                if isinstance(param.expiration, timedelta):
                    expiration_in_ns = param.expiration.total_seconds() * 1e9
                    param.expiration = commit_timestamp + expiration_in_ns

    @staticmethod
    def __generate_commit_id():
        random_string = "".join(
            SystemRandom().choice(string.printable) for _ in range(32)
        ).encode("utf-8")
        return hashlib.sha1(random_string).hexdigest()

    def __build_document(
        self, commit_id: Optional[str] = None, label: Optional[str] = None
    ) -> dict:
        """
        builds a document to be saved as a commit/temp in TinyDB.
        :param commit_id: is None when saving to temp.
        :param label: an optional label to associate the commit with.
        :return: a dictionary describing the current state of the ParamStore
        """
        metadata = self.__build_metadata(commit_id, label)
        if self.__is_in_memory_mode:
            params = copy.deepcopy(self.__params)
        else:
            params = self.__params
        return dict(metadata=metadata.__dict__, params=params, tags=self.__tags)

    @staticmethod
    def __build_metadata(
        commit_id: Optional[str] = None, label: Optional[str] = None
    ) -> Metadata:
        metadata = Metadata()
        metadata.id = commit_id
        metadata.timestamp = time.time_ns()
        metadata.label = label
        return metadata

    def checkout(
        self,
        commit_id: Optional[str] = None,
        commit_num: Optional[int] = None,
        move_by: Optional[int] = None,
    ) -> None:
        with self.__lock:
            commit = self.__get_commit(commit_id, commit_num, move_by)
            self.__params.clear()
            self.__params.update(commit["params"])
            self.__tags = commit["tags"]
            self.__base_commit_id = commit_id
            self.__base_doc_id = commit.doc_id
            self.__is_dirty = False
            self.__dirty_keys.clear()

    def list_commits(self, label: Optional[str] = None) -> List[Metadata]:
        with self.__lock:
            documents = self.__db.search(
                Query().metadata.label.test(_test_if_value_contains(label))
            )
            metadata = map(_extract_metadata, documents)
            return list(metadata)

    def __get_commit(
        self,
        commit_id: Optional[str] = None,
        commit_num: Optional[int] = None,
        move_by: Optional[int] = None,
    ) -> Document:
        if commit_id is not None:
            commit = self.__get_commit_by_id(commit_id)
        elif commit_num is not None:
            commit = self.__get_commit_by_num(commit_num)
        elif move_by is not None:
            commit = self.__get_commit_by_move_by(move_by)
        else:
            raise EntropyError(
                "Please provide one of the following arguments: "
                "commit_id, commit_num, or move_by"
            )
        return commit

    def __get_commit_by_id(self, commit_id: str):
        result = self.__db.search(Query().metadata.id == commit_id)
        if len(result) == 0:
            raise EntropyError(f"Commit with id '{commit_id}' not found")
        if len(result) > 1:
            raise EntropyError(
                f"{len(result)} commits with id '{commit_id}' found. "
                f"Only one commit is allowed per id"
            )
        return result[0]

    def __get_commit_by_num(self, commit_num: int):
        result = self.__db.get(doc_id=commit_num)
        if result is None:
            raise EntropyError(f"Commit with number '{commit_num}' not found")
        return result

    def __get_commit_by_move_by(self, move_by: int):
        doc_id = self.__base_doc_id + move_by
        result = self.__db.get(doc_id=doc_id)
        return result

    """ Merge """

    def merge(
        self,
        theirs: ParamStore,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.OURS,
    ) -> None:
        with self.__lock:
            ours = self
            self.__merge_trees(ours, theirs, merge_strategy)

    def __merge_trees(
        self,
        a: ParamStore | Dict,
        b: ParamStore | Dict,
        merge_strategy: MergeStrategy,
    ) -> bool:
        """Merges `b` into `a` *in-place* using the given strategy"""
        a_has_changed = False
        for key in b.keys():
            if key in a.keys():
                if (
                    (not isinstance(a[key], Param))
                    and isinstance(a[key], dict)
                    and (not isinstance(b[key], Param))
                    and isinstance(b[key], dict)
                ):
                    """This is a special case where the values of the Params are both
                    dictionaries. In this case we merge the dictionary from b into the
                    dictionary from a using the given strategy."""
                    a_has_changed = a_has_changed or self.__merge_trees(
                        a[key], b[key], merge_strategy
                    )
                    if (
                        a_has_changed
                        and isinstance(a, ParamStore)
                        and isinstance(b, ParamStore)
                    ):
                        """if the dictionary in a has been changed, this was done
                        in-place. We therefore need mark the Param key as dirty. We only
                        do this at the very top of the recursion - when a and b are the
                        ParamStores being merged"""
                        self.__is_dirty = True
                        self.__dirty_keys.add(key)
                elif a[key] == b[key]:
                    pass  # same leaf values, nothing to do
                else:  # diff leave values => conflict:
                    if merge_strategy == MergeStrategy.OURS:
                        pass  # a takes precedence, ignore b
                    elif merge_strategy == MergeStrategy.THEIRS:
                        a[key] = b[key]  # b takes precedence, overwrite a
                        a_has_changed = True
                    else:
                        raise NotImplementedError(
                            f"MergeStrategy '{merge_strategy}' is not implemented"
                        )
            else:  # key from b is not in a:
                a[key] = b[key]  # "copy" from b to a
                a_has_changed = True
        return a_has_changed

    def list_values(self, key: str) -> pd.DataFrame:
        with self.__lock:
            values = []
            commits = self.__db.all()
            commits.sort(key=lambda x: x["metadata"]["timestamp"])
            for commit in commits:
                try:
                    value = (
                        commit["params"][key].value,
                        _ns_to_datetime(commit["metadata"]["timestamp"]),
                        commit["metadata"]["id"],
                        commit["metadata"]["label"],
                    )
                    values.append(value)
                except KeyError:
                    pass
            if key in self.__params.keys():
                values.append((self[key], None, None, None))
            df = pd.DataFrame(values)
            if not df.empty:
                df.columns = ["value", "time", "commit_id", "label"]
            return df

    """ Tags """

    def add_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if key not in self.__params.keys():
                raise KeyError(f"key '{key}' is not in store")
            if tag not in self.__tags:
                self.__tags[tag] = []
            self.__tags[tag].append(key)
            self.__is_dirty = True

    def remove_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if tag not in self.__tags:
                return
            if key not in self.__tags[tag]:
                return
            self.__tags[tag].remove(key)
            self.__is_dirty = True

    def list_keys_for_tag(self, tag: str) -> List[str]:
        with self.__lock:
            if tag not in self.__tags:
                return []
            else:
                return self.__tags[tag]

    def list_tags_for_key(self, key: str):
        tags_for_key = []
        for item in self.__tags.items():
            if key in item[1]:
                tags_for_key.append(item[0])
        return tags_for_key

    """ Temporary State """

    def save_temp(self) -> None:
        """
        Saves the state of params to a temporary location
        """
        with self.__lock:
            table = self.__db.table(TEMP_TABLE)
            doc = self.__build_document()
            table.upsert(Document(doc, doc_id=1))

    def load_temp(self) -> None:
        """
        Overwrites the current state of params with data loaded from the temporary
        location
        """
        with self.__lock:
            table = self.__db.table(TEMP_TABLE)
            doc = table.get(doc_id=1)
            if not doc:
                raise EntropyError(
                    "Temp is empty. Call save_temp() before calling load_temp()"
                )
            self.__params.clear()
            self.__params.update(doc["params"])
            self.__tags = doc["tags"]


""" Static helper methods """


def _test_if_value_contains(label: str) -> Callable:
    return lambda val: (label or "") in (val or "")


def _extract_metadata(document: Document) -> Metadata:
    return Metadata(document.get("metadata"))


def _dict_to_json(d: Dict) -> str:
    return json.dumps(d, default=_json_dumps_default, sort_keys=True, ensure_ascii=True)


def _json_dumps_default(value):
    if isinstance(value, datetime):
        return str(value)
    else:
        return value.__dict__


def _ns_to_datetime(ns: int) -> pd.datetime:
    """Convert a UNIX epoch timestamp in nano-seconds to a human readable string"""
    return pd.to_datetime(ns)


def _map_dict(f, d: Dict) -> Dict:
    values_dict = dict()
    for item in d.items():
        k = item[0]
        v = item[1]
        if not isinstance(v, Param) and isinstance(v, dict):
            values_dict[k] = _map_dict(f, v)
        else:
            values_dict[k] = f(v)
    return values_dict


def _extract_param_values(d: Dict) -> Dict:
    return _map_dict(lambda x: x.value, d)


def _map_dict_in_place(f, d: Dict):
    for item in d.items():
        k = item[0]
        v = item[1]
        if isinstance(v, dict):
            _map_dict_in_place(f, v)
        else:
            d[k] = f(v)


def _extract_param_values_in_place(d: Dict):
    _map_dict_in_place(lambda x: x.value, d)


""" Migrations """


def migrate_param_store_0_1_to_0_2(path: str) -> None:
    """
    Backup and migrate an InProcessParamStore JSON file from storing values to storing
    Params containing the values. Preserves commits, timestamps and ids.

    :param path: path to an existing JSON TinyDB file containing params.
    """
    old_version = "0.1"
    new_version = "0.2"

    _check_version(path, old_version, new_version)
    backup_path = _backup_file(path)
    with TinyDB(backup_path) as old_db:
        old_commits = old_db.all()
        old_temp = old_db.table(TEMP_TABLE).get(doc_id=TEMP_DOC_ID)
    with TinyDB(path, storage=JSONPickleStorage) as new_db:
        for old_commit in old_commits:
            new_commit = copy.deepcopy(old_commit)
            _wrap_params(new_commit)
            _rename_ns(new_commit)
            new_db.insert(new_commit)
        if old_temp:
            new_temp = copy.deepcopy(old_temp)
            _wrap_params(new_temp)
            _rename_ns(new_temp)
            new_db.table(TEMP_TABLE).insert(new_temp)
        _set_version(new_db, new_version)


def _backup_file(path):
    backup_path = path.replace(".json", ".bak.json")
    shutil.move(path, backup_path)
    return backup_path


def _wrap_params(new_commit):
    params = new_commit["params"]
    for key in params.keys():
        value = params[key]
        if not isinstance(value, Param):
            params[key] = Param(value)
    # _map_dict_in_place(lambda val: Param(val), new_doc["params"])
    new_commit["params"] = params


def _rename_ns(new_commit):
    timestamp = new_commit["metadata"]["ns"]
    del new_commit["metadata"]["ns"]
    new_commit["metadata"]["timestamp"] = timestamp


def _set_version(db: TinyDB, version: str):
    info_table = db.table(INFO_TABLE)
    info_doc = info_table.get(doc_id=INFO_DOC_ID)
    if info_doc:
        info_doc[VERSION_KEY] = version
    else:
        info_table.insert(dict(VERSION_KEY=version))


def _get_version(db: TinyDB):
    if INFO_TABLE in db.tables():
        info_table = db.table(INFO_TABLE)
        info_doc = info_table.get(doc_id=INFO_DOC_ID)
        return info_doc[VERSION_KEY]
    return "0.1"


def _check_version(path: str, old_version: str, new_version: str):
    if os.path.isfile(path):
        with TinyDB(path) as old_db:
            actual_version = _get_version(old_db)
            if actual_version != old_version:
                raise EntropyError(
                    f"Cannot migrate file '{path}' from version {actual_version} to "
                    f"version {new_version}"
                )
    else:
        raise EntropyError(
            f"Cannot migrate file '{path}' to version {new_version} because the file "
            "does not exist"
        )



from entropylab.pipeline.api.data_reader import ExperimentReader
from entropylab.pipeline.api.data_writer import RawResultData
from entropylab.pipeline.api.execution import EntropyContext
from entropylab.pipeline.api.graph import GraphHelper
from entropylab.pipeline.graph_experiment import (
    Graph,
    PyNode,
    SubGraphNode,
    pynode,
)
from entropylab.components.lab_topology import ExperimentResources, LabResources
from entropylab.pipeline.results_backend.sqlalchemy.db import SqlAlchemyDB
from entropylab.pipeline.script_experiment import Script, script_experiment
from entropylab.pipeline.api.in_process_param_store import InProcessParamStore
from entropylab.quam.core import QuAMManager, QuAM

__all__ = [
    "ExperimentReader",
    "RawResultData",
    "EntropyContext",
    "GraphHelper",
    "Graph",
    "PyNode",
    "SubGraphNode",
    "pynode",
    "ExperimentResources",
    "LabResources",
    "SqlAlchemyDB",
    "Script",
    "script_experiment",
    "InProcessParamStore",
    "QuAM",
    "QuAMManager",
]



"""simplifying node id

Revision ID: 9ffd2ba0d5bf
Revises: f1ada2484fe2
Create Date: 2022-04-10 08:26:35.076466+00:00

"""
from alembic import op

# from sqlalchemy import Column, Integer

# revision identifiers, used by Alembic.
revision = "9ffd2ba0d5bf"
down_revision = "f1ada2484fe2"
branch_labels = None
depends_on = None

NodeTable = "Nodes"


def upgrade():
    op.execute("""DROP TABLE IF EXISTS "Nodes_new";""")

    op.execute(
        """CREATE TABLE "Nodes_new" (
        id INTEGER PRIMARY KEY,
        experiment_id INTEGER NOT NULL,
        stage_id INTEGER NOT NULL,
        label VARCHAR,
        start DATETIME NOT NULL,
        is_key_node BOOLEAN,
        FOREIGN KEY(experiment_id) REFERENCES "Experiments" (id) ON DELETE CASCADE);"""
    )

    op.execute(
        """INSERT INTO Nodes_new
        (experiment_id, stage_id, label, start, is_key_node)
        SELECT experiment_id, id, label, start, is_key_node
        FROM Nodes;"""
    )

    op.execute("""ALTER TABLE "Nodes" RENAME TO "Nodes_old";""")

    op.execute("""ALTER TABLE "Nodes_new" RENAME TO "Nodes";""")


def downgrade():
    op.execute("""ALTER TABLE "Nodes" RENAME TO "Nodes_new";""")

    op.execute("""ALTER TABLE "Nodes_old" RENAME TO "Nodes";""")

    pass



from . import db
from .db import SqlAlchemyDB

from entropylab.pipeline.results_backend.sqlalchemy.db_initializer import (
    _DbInitializer,
    _DbUpgrader,
)


def init_db(path: str):
    SqlAlchemyDB(path)


def upgrade_db(path: str):
    """Upgrades an Entropy SQLite database to the latest schema version

    * Be sure to back up your database to a safe place before upgrading it *.

    :param path: The path to the SQLite database to be upgraded
    """
    _DbUpgrader(path).upgrade_db()



# ==================== DEFINE NODE ====================
import entropylab.flame.nodeio as nodeio

nodeio.context(
    name="CheerfulNode",
    description="greets customers",
    icon="bootstrap/person-circle.svg",
)

inputs = nodeio.Inputs()
inputs.stream("customer", units="human", description="one person at a time")
inputs.state("weather", units="best guess", description="How is weather today")

outputs = nodeio.Outputs()
outputs.define(
    "requested_salary",
    units="k$",
    description="requested fees from administrator",
    retention=2,
)

nodeio.register()

# ==================== DRY RUN DATA ====================

inputs.set(weather="sunny")
inputs.set(customer="Alice")
inputs.set(customer="Bob")
inputs.set(customer="Mars")
inputs.set(customer="Venus")

# =============== RUN NODE STATE MACHINE ===============

part_of_day = 0
salary_demand = 0.0

while nodeio.status.active:
    person = inputs.get("customer")
    weather = inputs.get("weather")
    if part_of_day % 2 == 0:
        day_time = "morning"
    else:
        day_time = "afternoon"
    print(f"Hi {person}, it is a {weather} {day_time}")
    part_of_day += 1

    salary_demand += 1.2

    outputs.set(requested_salary=salary_demand)



from dash import dash_table

from entropylab.dashboard.theme import (
    table_style_data,
    table_style_filter,
    table_style_header,
    table_style_cell,
)


def table(records):
    tbl = dash_table.DataTable(
        id="experiments-table",
        columns=[
            dict(name="id", id="id", type="numeric"),
            dict(name="label", id="label", type="text"),
            dict(name="start_time", id="start_time", type="datetime"),
            dict(name="end_time", id="end_time", type="datetime"),
            dict(name="user", id="user", type="text"),
            # dict(name="success", id="success"),
        ],
        data=records,
        persistence=True,
        persistence_type="session",
        row_selectable="multi",
        cell_selectable=False,
        sort_action="native",
        filter_action="native",
        style_data=table_style_data,
        style_filter=table_style_filter,
        style_header=table_style_header,
        style_cell=table_style_cell,
        style_cell_conditional=[
            {"if": {"column_id": "id"}, "width": "9%"},
            {"if": {"column_id": "label"}, "width": "20%"},
            {"if": {"column_id": "start_time"}, "width": "26%"},
            {"if": {"column_id": "end_time"}, "width": "26%"},
            {"if": {"column_id": "user"}, "width": "13%"},
            {"if": {"column_id": "success"}, "width": "6%"},
        ],
        style_data_conditional=[
            {
                "if": {
                    "column_id": "success",
                },
                "textAlign": "center",
            }
        ],
        # tooltip_data=[
        #     {
        #         column: {"value": str(value), "type": "markdown"}
        #         for column, value in row.items()
        #     }
        #     for row in records
        # ],
    )

    return tbl



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["CalibrateSystem"]


class CalibrateSystem(object):
    def __init__(self, workflow_node_unique_name, instrumentAddress=None):
        """Makes sure that system is calibrated

        :param instrumentAddress: (IP address - STATE) instrument IP address for calibration
        """
        self._command = "python3"
        self._bin = "calibrate.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/gear-fill.svg"
        self._inputs = _Inputs(
            instrumentAddress=instrumentAddress,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, instrumentAddress=None):
        self._inputs = Inputs()

        self._inputs.state(
            "instrumentAddress",
            description="instrument IP address for calibration",
            units="IP address",
        )
        self._inputs.set(instrumentAddress=instrumentAddress)

    @property
    def instrumentAddress(self):
        """Input: instrument IP address for calibration (IP address)"""
        return self._inputs.get("instrumentAddress")

    @instrumentAddress.setter
    def instrumentAddress(self, value):
        """Input: instrument IP address for calibration (IP address)"""
        self._inputs.set(instrumentAddress=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "calibration_status",
        ]

    @property
    def calibration_status(self):
        """Output: Calibration status
        :return: (calibrated, uncalibrated)
        """
        return "#" + self._name + "/calibration_status"



import inspect
import sys
import traceback
from inspect import signature
from typing import Callable, Any
from typing import Optional

from entropylab.pipeline.api.data_reader import ExperimentReader
from entropylab.pipeline.api.data_writer import DataWriter
from entropylab.pipeline.api.execution import (
    ExperimentExecutor,
    EntropyContext,
    _EntropyContextFactory,
)
from entropylab.pipeline.api.experiment import (
    ExperimentDefinition,
    ExperimentHandle,
    _Experiment,
)
from entropylab.components.lab_topology import ExperimentResources
from entropylab.logger import logger


def script_experiment(
    label: str, resources: ExperimentResources = None, db: Optional[DataWriter] = None
):
    """
        decorator for running a script experiment with the given python function
    :param label: experiment label
    :param resources: experiment resources if used in the function
    :param db: results db if used in the function
    :return: the handle of experiment
    """

    def decorate(fn):
        Script(resources, fn, label).run(db)

    return decorate


class _ScriptExecutor(ExperimentExecutor):
    def __init__(self, script: Callable) -> None:
        super().__init__()
        self._script: Callable = script
        self._stopped = False

    def execute(self, entropy_context_factory: _EntropyContextFactory) -> Any:
        try:
            context = entropy_context_factory.create()
            sig = signature(self._script)
            keyword_function_parameters = {}
            for param in sig.parameters:
                if sig.parameters[param].annotation is EntropyContext:
                    keyword_function_parameters[param] = context

            if len(sig.parameters) > 0:
                return self._script(**keyword_function_parameters)
            else:
                return self._script()
        except BaseException as e:
            self._stopped = True
            trace = "\n".join(traceback.format_exception(*sys.exc_info()))
            logger.error(
                f"Stopping Script, Error message: {e} of type "
                f"{e.__class__.__qualname__}.\ntrace:\n{trace}"
            )
            return

    @property
    def failed(self) -> bool:
        return self._stopped


class ScriptExperimentHandle(ExperimentHandle):
    """
    An handle of the script experiment execution
    can be used to get information and read results
    """

    def __init__(self, experiment: _Experiment) -> None:
        super().__init__()
        self._experiment = experiment

    @property
    def id(self):
        return self._experiment.exp_id

    @property
    def results(self) -> ExperimentReader:
        return ExperimentReader(self.id, self._experiment.data_reader())


class Script(ExperimentDefinition):
    """
    Definition of a Script Experiment that gets a python function and run it within Entropy.
    This is the most simple type of experiment, which still handle resources and results,
        but using a simple function.
    """

    def __init__(
        self,
        resources: Optional[ExperimentResources],
        script: Callable,
        label: Optional[str] = None,
        story: str = None,
    ) -> None:
        """
            Script Experiment that gets a python function and run it within Entropy.
        :param resources: shared lab resources or temporary resources
                        that are used in the experiment.
        :param script: the experiment itself in a python function
        :param label: experiment label
        :param story: a description of the experiment, which will create an experiment story
                         with all other parts of the experiment
        """
        super().__init__(resources, label, story)
        self._script = script

    def _get_execution_instructions(self) -> ExperimentExecutor:
        return _ScriptExecutor(self._script)

    def run(self, db: Optional[DataWriter] = None, **kwargs) -> ScriptExperimentHandle:
        experiment = self._run(db, **kwargs)
        return ScriptExperimentHandle(experiment)

    def serialize(self) -> str:
        """
        source code of the given python function
        """
        return inspect.getsource(self._script)



import dash_bootstrap_components as dbc
from dash import html, dcc

from entropylab.dashboard.pages.components.footer import footer
from entropylab.dashboard.pages.components.top_bar import top_bar
from entropylab.dashboard.pages.results.dashboard_data import (
    DashboardDataReader,
)
from entropylab.dashboard.pages.results.table import table
from entropylab.pipeline.results_backend.sqlalchemy.project import (
    project_name,
)

REFRESH_INTERVAL_IN_MILLIS = 3000


def build_layout(path: str, dashboard_data_reader: DashboardDataReader):
    records = dashboard_data_reader.get_last_experiments()

    return (
        dbc.Container(
            fluid=True,
            className="main",
            children=[
                dcc.Store(id="figures-by-key", storage_type="session"),
                dcc.Store(id="plot-keys-to-combine", storage_type="session"),
                dcc.Store(id="prev-selected-rows", storage_type="session"),
                dcc.Interval(
                    id="interval", interval=REFRESH_INTERVAL_IN_MILLIS, n_intervals=0
                ),
                dbc.Alert(
                    "",
                    id="failed-plotting-alert",
                    color="warning",
                    is_open=False,
                    fade=True,
                ),
                dbc.Modal(
                    [
                        dbc.ModalHeader(
                            dbc.ModalTitle(
                                f"ℹ️ Project '{project_name(path)}' contains no "
                                f"experiments"
                            ),
                            close_button=False,
                        ),
                        dbc.ModalBody(
                            "As soon as experiments are saved to the project this "
                            "notice will be closed and the experiments will be shown "
                            "below."
                        ),
                        dbc.ModalFooter(),
                    ],
                    id="empty-project-modal",
                    backdrop="static",
                    centered=True,
                    is_open=False,
                ),
                top_bar(path),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                html.H5("Experiments", id="experiments-title"),
                                (table(records)),
                            ],
                            width="5",
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    [
                                        html.H5("Plots and Figures", id="plots-title"),
                                        dcc.Loading(
                                            id="plot-tabs-loading",
                                            children=[dbc.Tabs(id="plot-tabs")],
                                            type="default",
                                        ),
                                    ]
                                ),
                                dbc.Row(
                                    dbc.Button(
                                        "➕ Add Plot to Aggregate View",
                                        id="add-button",
                                    ),
                                    className="add-button-container",
                                ),
                                dbc.Row(
                                    [
                                        html.H5(
                                            "Aggregate View", id="experiments-title"
                                        ),
                                        dbc.Tabs(
                                            id="aggregate-tabs",
                                            children=[
                                                dbc.Tab(
                                                    "",
                                                    label="",
                                                    id="aggregate-tab",
                                                )
                                            ],
                                            persistence=True,
                                        ),
                                        dbc.Button(
                                            [
                                                " Copy Data to Clipboard",
                                                dcc.Clipboard(
                                                    title="Copy data to clipboard",
                                                    id="aggregate-clipboard",
                                                    className="position-absolute "
                                                    # "start-0 top-0 "
                                                    # "h-100 w-100 opacity-0",
                                                ),
                                            ],
                                            id="copy-data-button",
                                            color="primary",
                                        ),
                                    ],
                                    id="aggregate-container",
                                ),
                                dbc.Row(
                                    [
                                        html.Div(
                                            "Click to remove from Aggregate View:",
                                            id="remove-title",
                                        ),
                                        html.Div(id="remove-buttons"),
                                    ],
                                ),
                            ],
                            width="7",
                        ),
                    ],
                ),
                footer(),
            ],
        ),
    )



"""wrapping_param_store_values

Revision ID: 06140c96c8c4
Revises: 9ffd2ba0d5bf
Create Date: 2022-05-19 09:12:01.569825+00:00

"""

import logging
import os
import shutil

from alembic import op

from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.api.in_process_param_store import (
    migrate_param_store_0_1_to_0_2,
)
from entropylab.pipeline.results_backend.sqlalchemy.project import param_store_file_path

# revision identifiers, used by Alembic.
revision = "06140c96c8c4"
down_revision = "9ffd2ba0d5bf"
branch_labels = None
depends_on = None


def upgrade():
    logger = logging.getLogger(__name__)
    conn = op.get_bind()
    project_path = os.path.abspath(
        os.path.join(os.path.dirname(conn.engine.url.database), "..")
    )
    json_file_path = param_store_file_path(project_path)
    logger.debug(
        f"Attempting to migrate InProcessParamStore file {json_file_path} from v0.1 to "
        f"v0.2"
    )
    try:
        migrate_param_store_0_1_to_0_2(str(json_file_path))
    except EntropyError as ee:
        logger.warning(str(ee))


def downgrade():
    logger = logging.getLogger(__name__)
    conn = op.get_bind()
    project_path = os.path.abspath(
        os.path.join(os.path.dirname(conn.engine.url.database), "..")
    )
    json_file_path = str(param_store_file_path(project_path))
    backup_json_file_path = json_file_path.replace(".json", ".bak.json")
    downgraded_json_file_path = json_file_path.replace(".json", ".downgraded.json")
    logger.debug(
        f"Attempting to downgrade  InProcessParamStore file {json_file_path} from v0.2 "
        f"to v0.1 using backup file {backup_json_file_path}"
    )
    shutil.move(json_file_path, downgraded_json_file_path)
    shutil.move(backup_json_file_path, json_file_path)
    logger.debug(
        f"Attempting to migrate InProcessParamStore file {json_file_path} from v0.1 to "
        f"v0.2"
    )



from dynaconf import Dynaconf

settings = Dynaconf(
    envvar_prefix="ENTROPY",
    settings_files=["settings.toml", ".secrets.toml"],
)

# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load this files in the order.



from sqlalchemy import (
    Column,
    Integer,
    String,
    DATETIME,
    ForeignKey,
    Enum,
    Boolean,
)

from entropylab.components.lab_topology import (
    DriverType,
)
from entropylab.pipeline.results_backend.sqlalchemy.model import Base


class Resources(Base):
    __tablename__ = "Resources"

    id = Column(Integer, primary_key=True)
    update_time = Column(DATETIME, nullable=False)
    name = Column(String)
    driver = Column(String)
    module = Column(String)
    class_name = Column(String)
    version = Column(String)
    driver_type = Column(Enum(DriverType))
    args = Column(String)
    kwargs = Column(String)
    deleted = Column(Boolean, default=False)
    number_of_experiment_args = Column(Integer)
    keys_of_experiment_kwargs = Column(String)
    cached_metadata = Column(String)


class ResourcesSnapshots(Base):
    __tablename__ = "ResourcesSnapshots"

    id = Column(Integer, primary_key=True)
    update_time = Column(DATETIME, nullable=False)
    name = Column(String)
    driver_id = Column(Integer, ForeignKey("Resources.id", ondelete="CASCADE"))
    state = Column(String)



import argparse
import functools
import sys

import pkg_resources

from entropylab.dashboard import serve_dashboard
from entropylab.logger import logger
from entropylab.pipeline.results_backend.sqlalchemy import init_db, upgrade_db


# Decorator for friendly error messages


def command(func: callable) -> callable:
    """Decorator that runs commands. On error, prints friendly message when possible"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except RuntimeError as re:
            command_name = func.__name__
            logger.exception(
                "RuntimeError in Entropy CLI command %s, args: %s", command_name, args
            )
            print(re, file=sys.stderr)
            sys.exit(-1)

    return wrapper


# CLI command functions


@command
def init(args: argparse.Namespace):
    init_db(args.directory)


@command
def upgrade(args: argparse.Namespace):
    upgrade_db(args.directory)


@command
def serve(args: argparse.Namespace):
    serve_dashboard(args.directory, args.host, args.port, args.debug)


# The parser


def _build_parser():
    parser = argparse.ArgumentParser()
    # in case no arguments were supplied:
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers()

    directory_arg = {
        "help": "path to a directory containing Entropy project",
        "nargs": "?",
        "default": ".",
    }

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s " + pkg_resources.get_distribution("entropylab").version,
    )

    # init
    init_parser = subparsers.add_parser("init", help="initialize a new Entropy project")
    init_parser.add_argument("directory", **directory_arg)
    init_parser.set_defaults(func=init)

    # upgrade
    upgrade_parser = subparsers.add_parser(
        "upgrade", help="upgrade an Entropy project to the latest version"
    )
    upgrade_parser.add_argument("directory", **directory_arg)
    upgrade_parser.set_defaults(func=upgrade)

    # serve
    serve_parser = subparsers.add_parser(
        "serve", help="serve & launch the results dashboard app in a browser"
    )
    serve_parser.add_argument("directory", **directory_arg)
    serve_parser.add_argument(
        "host",
        help="host name from which to serve the dashboard app",
        nargs="?",
        default=None,
    )
    serve_parser.add_argument(
        "port",
        help="port number from which to serve the dashboard app",
        nargs="?",
        default=None,
    )
    serve_parser.add_argument("--debug", dest="debug", action="store_true")
    serve_parser.set_defaults(func=serve, debug=False)

    return parser


# main


def main():
    parser = _build_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# Flame executor

import signal

from entropylab.flame.execute import _utils as execute_utils
from entropylab.flame.execute._config import logger
from entropylab.flame.execute._execute import Execute


def _main():
    logger.info("Execute. Start")
    signal.signal(signal.SIGINT, execute_utils.exit_gracefully)
    signal.signal(signal.SIGTERM, execute_utils.exit_gracefully)
    args = execute_utils.parse_args()
    execute_ = Execute(args)
    msg = execute_.run()
    print(msg)
    logger.info("Execute. Finish")


if __name__ == "__main__":
    _main()



import json
import socket
import argparse

import zmq
import msgpack
import contextlib

from tzlocal import get_localzone
from datetime import datetime, timezone

from sqlalchemy import create_engine

from entropylab.flame.execute._config import _Config, logger


def is_port_in_use(port, host="localhost"):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        in_use = s.connect_ex((host, port)) == 0
        logger.debug(f"Utils. Check port {port} in use. Result: {in_use}. Host: {host}")
        return in_use


def status_update(node_name, routing_key, message, style, updates_channel):
    logger.debug(
        f"Utils. Status update. Node name: {node_name}; "
        f"routing key: {routing_key}; message: {message}; "
        f"style: {style}"
    )
    if updates_channel is not None:
        update = {"node": node_name, "msg": message, "style": style}
        update = json.dumps(update)
        updates_channel.basic_publish(
            exchange="amq.topic", routing_key=routing_key, body=update.encode()
        )
    _Config.node_status_dict[node_name] = message


def get_free_port(start_port_number, runtime_state, eui, end_port_number=64000):
    logger.debug(f"Utils. Get free port. Start port: {start_port_number}. EUI: {eui}")
    port_in_use = True
    port_number = start_port_number
    while port_in_use and port_number < end_port_number:
        port_number += 1
        # set lock by entropy if possible on this port
        if runtime_state.setnx(f"system/port{port_number}", eui):
            # lock set, is the port in use?
            port_in_use = is_port_in_use(port_number)
            if port_in_use:
                # port is in use
                # clear lock by entropy
                runtime_state.delete(f"system/port{port_number}")
    if port_number == end_port_number:
        raise ValueError("No free ports for connecting nodes found")
    logger.debug(f"Utils. Get free port. Founded port: {port_number}")
    return port_number


def remove_port_lock(port_number, runtime_state):
    logger.debug(f"Utils. Remove port lock. Port: {port_number}")
    runtime_state.delete(f"system/port{port_number}")


def exit_gracefully(self, *args):
    logger.debug("Utils. Exit gracefully")
    _Config.execution_active = False


def check_node_messages(executor_input):
    logger.debug("Utils. Check node messages.")
    try:
        # check for updates
        msg = msgpack.unpackb(executor_input.recv(flags=zmq.NOBLOCK))
        logger.debug(f"Utils. Check node messages. Message: {msg}")
        if msg["status"] == "requests workflow termination":
            _Config.execution_active = False
        return
    except zmq.error.ZMQError as e:
        if e.errno == zmq.EAGAIN:
            # the state value has not been updated since last
            # get call
            return
        else:
            logger.error(f"Utils. Check node messages. Exception: {str(e)}")
            return


@contextlib.contextmanager
def get_db_manager(db_url):
    engine = create_engine(db_url)  # SQLALCHEMY_DATABASE_URL
    conn = engine.connect()
    try:
        yield conn
    finally:
        conn.close()


@contextlib.contextmanager
def get_runtimedata(db_name):
    engine = create_engine(db_name)  # DATABASE_NAME
    conn = engine.connect()
    try:
        yield conn
    finally:
        conn.close()


def write_metadata_to_h5(h5file, metadata):
    logger.debug(f"Utils. Write metadata to h5 file. Metadata: {metadata}")
    h5file.attrs["project"] = metadata.get("project", "")
    h5file.attrs["job_eui"] = metadata.get("job_eui", "")
    h5file.attrs["job_description"] = metadata.get("job_description", "")
    h5file.attrs["workflow_eui"] = metadata.get("workflow_eui", "")
    h5file.attrs["parameters_eui"] = metadata.get("parameters_eui", "")
    if "workflow_commit_id" in metadata:
        h5file.attrs["workflow_commit"] = metadata.get("workflow_commit_id")
    if "parameters_commit_id" in metadata:
        h5file.attrs["parameters_commit"] = metadata.get("parameters_commit_id")

    utc_dt = datetime.now(timezone.utc)
    h5file.attrs["creation_timestamp"] = "{}".format(
        utc_dt.astimezone(get_localzone()).isoformat()
    )


def parse_args():
    parser = argparse.ArgumentParser(
        description="Flame executor of parametrized workflows"
    )
    parser.add_argument(
        "-w",
        "--workflow",
        type=str,
        default="workflow.py",
        help="Python file that defines main entropylab.Workflow (default workflow.py)",
    )
    parser.add_argument(
        "-p",
        "--parameters",
        type=str,
        default="parameters.json",
        help="JSON file that resolves workflow parameters",
    )
    parser.add_argument(
        "-t",
        "--max-execution-time",
        type=int,
        default=60,
        help="Maximal execution time in s",
    )
    parser.add_argument(
        "-d",
        "--status-check-interval",
        type=int,
        default=1,
        help="Node status check interval in s (default 1 s)",
    )
    parser.add_argument(
        "-m",
        "--metadata",
        type=str,
        default='{"project": "test_from_command_line", '
        '"job_description": "Flame execution from command line"}',
        help="Metadata about the particular executed job in json string format consisting of "
        "project prefix, runtime id, job eui, workflow eui, parameters eui, "
        "workflow commit hash, parameters commit hash, job description. For example: "
        '\'{"project": "default_project", "runtime_id": 1, "job_eui": "#/j1", '
        '"workflow_eui": "#/w1", "parameters_eui": "#/p1", '
        '"workflow_commit_id": "b3a8f6973ea45a1fc34bfd83e1e67b90066253eb", '
        '"parameters_commit_id": "9c41a22a6502ad4b26f63011f5701ac4c5ef1861", '
        '"job_description": "some description"}\'',
    )
    parser.add_argument(
        "-c",
        "--connection-wait",
        type=int,
        default=5,
        help="How long in seconds to wait for succesful establishment of "
        "communication between nodes before timeout.",
    )
    return parser.parse_args()



import os
import pathlib

import redis
import sqlalchemy

from sqlalchemy import text as sql_text

from entropylab.flame.execute._config import _Config, logger
from entropylab.flame.execute import _utils as execute_utils


def _setup_runtime_db(db_url, runtimedata_info):
    logger.debug("RuntimeStateInfo. Setup runtime database")
    with execute_utils.get_db_manager(db_url) as db:
        try:
            with db.begin():
                db.execute(sql_text("commit"))
                db.execute(
                    sql_text(
                        f"""
                SELECT pg_terminate_backend(pg_stat_activity.pid)
                FROM pg_stat_activity
                WHERE pg_stat_activity.datname = '{runtimedata_info[3]}'
                AND pid <> pg_backend_pid();
                """
                    )
                )
                db.execute(sql_text(f"DROP DATABASE {runtimedata_info[3]}"))
                db.execute(sql_text("DROP OWNED BY workflow"))
                db.execute(sql_text("DROP USER workflow"))
        except Exception:
            # no such database
            logger.info("RuntimeStateInfo. No such database")
            pass

        with db.begin():
            db.execute(sql_text("commit"))
            db.execute(
                sql_text(f"CREATE DATABASE {runtimedata_info[3]} ENCODING 'UTF8'")
            )
            db.execute(sql_text(f"CREATE USER {runtimedata_info[0]}"))
            db.execute(
                sql_text(
                    f"ALTER USER {runtimedata_info[0]} PASSWORD '{runtimedata_info[1]}'"
                )
            )
            db.execute(
                sql_text(
                    f"""GRANT ALL PRIVILEGES ON DATABASE
                    {runtimedata_info[3]} TO {runtimedata_info[0]}"""
                )
            )


def _setup_runtime_state(host_url: str = "localhost", port=5431):
    logger.debug("RuntimeStateInfo. Setup runtime state")
    playbook_server = [host_url, 6379, 0]
    # db_user, db_password, db_URL, db_name
    runtimedata_info = ["workflow", "datapass", f"{host_url}:{port}", "workflow"]
    runtime_state = redis.Redis(
        host=playbook_server[0], port=playbook_server[1], db=playbook_server[2]
    )
    runtime_state.ping()
    return runtime_state, playbook_server, runtimedata_info


class RuntimeStateInfo:
    def __init__(self):
        rs, ps, rd_i = None, None, None
        try:
            rs, ps, rd_i = _setup_runtime_state("localhost", 5431)
            logger.info("RuntimeStateInfo. Host: localhost")
        except redis.ConnectionError:
            rs, ps, rd_i = _setup_runtime_state("runtimedata", 5432)
            logger.info("RuntimeStateInfo. Host: runtimedata")
        except Exception as e:
            logger.error(
                f"Unexpected exception: {str(e)}.\nTraceback: {e.__traceback__}.\n"
                f"To run Flame executor you have to have EntropyHub runtime layer "
                f"running on the computer. Please double check that this is not missing."
            )
        self.runtime_state = rs
        self.playbook_server = ps
        self.runtimedata_info = rd_i
        data_server_arg = f"{rd_i[0]},{rd_i[1]},{rd_i[2]},{rd_i[3]}"
        rs.set("dataserver", data_server_arg)
        _Config.DATABASE_NAME = (
            f"postgresql://{rd_i[0]}:{rd_i[1]}"
            f'@{rd_i[2]}/{rd_i[3]}?client_encoding="utf8"'
        )
        admin_password = os.getenv("POSTGRES_PASSWORD", "passwordpasswordpassword")
        _Config.SQLALCHEMY_DATABASE_URL = (
            f"postgresql://postgres:{admin_password}"
            f'@{rd_i[2]}/postgres?client_encoding="utf8"'
        )
        _setup_runtime_db(_Config.SQLALCHEMY_DATABASE_URL, rd_i)
        self.install_extensions()

    @staticmethod
    def install_extensions():
        logger.debug("RuntimeStateInfo. Install extensions")
        with execute_utils.get_runtimedata(_Config.DATABASE_NAME) as db:
            with db.begin():
                db.execute(sql_text("CREATE EXTENSION IF NOT EXISTS timescaledb"))

                msgpack_sql_extension = os.path.join(
                    pathlib.Path(__file__).parent.resolve(), "./msgpack_decode.sql"
                )
                with open(msgpack_sql_extension) as file:
                    escaped_sql = sqlalchemy.text(file.read())
                    db.execute(escaped_sql)



from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Set, Dict, Any, List, Optional

import networkx as nx
from graphviz import Digraph

from entropylab.pipeline.api.execution import EntropyContext


@dataclass(frozen=True, eq=True)
class Output:
    """
    Node output respresentation that contain the node and the output name
    """

    node: Node
    name: str


@dataclass
class RetryBehavior:
    """
    Attributes:
        number_of_attempts: the maximum number of retries.
        wait_time: the initial wait time between each attempt [seconds].
        backoff: a factor that multiplies the delay on each attempt (1 is no backoff).
        added_delay: delay that is added on each attempt [seconds].
        max_wait_time: optional maximum delay.
    """

    number_of_attempts: float = 5
    wait_time: float = 10
    backoff: float = 2
    added_delay: float = 0
    max_wait_time: Optional[float] = None


class Node(ABC):
    """
    An abstract class for Entropy graph node.
    Node is defined by some functionality, input and outputs.
    Different implementations of node can be created, and connected one to the other.
    """

    def __init__(
        self,
        label: str = None,
        input_vars: Dict[str, Output] = None,
        output_vars: Set[str] = None,
        must_run_after: Set[Node] = None,
        save_results: bool = True,
        retry_on_error: RetryBehavior = None,
    ):
        """
            An abstract class for Entropy graph node.
        :param label: node label
        :param input_vars: dictionary of node inputs, keys are input names.
                        the input values should be defined as the following:
                            >>>input_vars={"a": node.outputs["x"]}
        :param output_vars: a set of outputs name. The given python function should
                        return a dictionary, which it's keys are the same as output vars
        :param must_run_after: A set of nodes. If those nodes are in the same graph,
                            current node will run after they finish execution.
        :param save_results: True to save the node outputs to results db.
        """
        self._label = label
        self._input_vars = input_vars
        if self._input_vars is None:
            self._input_vars = {}
        self._output_vars: Set[str] = output_vars
        if self._output_vars is None:
            self._output_vars = {}
        self._must_run_after: Set[Node] = must_run_after
        if self._must_run_after is None:
            self._must_run_after = {}
        self._save_results = save_results
        self._retry_on_error = retry_on_error

    @property
    def label(self) -> str:
        """
        :return: node label
        """
        return self._label

    @property
    def outputs(self) -> Dict[str, Output]:
        """
        :return: a dictionary of outputs representations, indexed by name
        """
        return {
            output_name: Output(self, output_name) for output_name in self._output_vars
        }

    def add_input(self, name: str, parent_node_output: Output):
        """
            adds a new input to current node
        :param name: input name
        :param parent_node_output: the parent node output representation
        """
        if name in self._input_vars:
            raise KeyError(f"Input {name} already exist")

        self._input_vars[name] = parent_node_output

    def get_parents(self) -> List[Node]:
        """
        :return: list of all parent nodes
        """
        return list(set([var.node for var in self._input_vars.values()])) + list(
            self._must_run_after
        )

    def get_inputs(self) -> List[Output]:
        """
        :return: list of all input, as the parent node output representations
        """
        return list(self._input_vars.values())

    def get_inputs_by_name(self) -> Dict[str, Output]:
        """
        :return: a dictionary of node inputs as the parent node output representation,
                indexed by name
        """
        return self._input_vars

    @abstractmethod
    def _execute(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        """
            Execute the node - run the node function with the given input and context.
            The abstract method _execute should be implemented by all subclasses.
        :param input_values: a dictionary of inputs, indexed by input name
        :param context: Entropy context of the specific node.
        :param is_last: True if this is the last node of the graph.
        :param kwargs: extra key word arguments passed by the user in definition.run function
        """
        pass

    @abstractmethod
    async def _execute_async(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        """
            Execute the node - run the node function with the given input and context.
            The abstract method _execute_async should be implemented by all subclasses.
            the method is an async method, so it can run in asyncio context.
        :param input_values: a dictionary of inputs, indexed by input name
        :param context: Entropy context of the specific node.
        :param is_last: True if this is the last node of the graph.
        :param kwargs: extra key word arguments passed by the user in definition.run function
        """
        pass

    def ancestors(self) -> Set[Node]:
        """
        :return: a set of node's ancestors, including current node
        """
        parents: Set[Node] = set(
            [var.node for var in self._input_vars.values()] + list(self._must_run_after)
        )
        ancestors = parents.copy()
        ancestors.add(self)
        for anc in parents:
            ancestors.update(anc.ancestors())
        return ancestors

    def _should_save_results(self):
        return self._save_results

    def _retry_on_error_function(self) -> RetryBehavior:
        return self._retry_on_error


@dataclass(frozen=True, eq=True)
class _NodeExecutionInfo:
    node: Node
    is_key_node: bool


class GraphHelper:
    """
    Class representing a graph, with relevant graph functions and algorithms
    """

    def __init__(self, nodes: Set[_NodeExecutionInfo]) -> None:
        """
            Class representing a graph, with relevant graph functions and algorithms
        :param nodes: complete set of nodes that assembles the graph
        """
        super().__init__()
        self._nodes: Set[_NodeExecutionInfo] = nodes

    @property
    def nodes(self) -> Set[Node]:
        """
            a set of all graph nodes
        :return:
        """
        return set([node.node for node in self._nodes])

    @property
    def leaves(self) -> Set[Node]:
        """
            a set of graph leaves
        :return:
        """
        end_nodes = {node.node for node in self._nodes}
        for node in self._nodes:
            for parent in node.node.get_parents():
                if parent in end_nodes:
                    end_nodes.remove(parent)
        return end_nodes

    def export_dot_graph(self) -> Digraph:
        """
        Converts the graph into DOT graph format
        :return: str
        """
        visited_labels = {}
        node_labels = {}

        def unique_label(node):
            if node in node_labels:
                return node_labels[node]
            if node.label in visited_labels:
                index = visited_labels[node.label] + 1
                visited_labels[node.label] = index
                node_labels[node] = f"{node.label}_{str(index)}"
            else:
                visited_labels[node.label] = 1
                node_labels[node] = node.label
            return node_labels[node]

        dot = Digraph()
        nodes = self._nodes
        for node in nodes:
            dot.node(unique_label(node.node), shape="box")

        for node in nodes:
            parents = node.node.get_parents()
            inputs = node.node.get_inputs()
            for parent in parents:
                input_names = [
                    input.name
                    for input in filter(lambda input: input.node == parent, inputs)
                ]
                dot.edge(
                    unique_label(parent), unique_label(node.node), ",".join(input_names)
                )

        dot.graph_attr["rankdir"] = "LR"
        return dot

    def nodes_in_topological_order(self) -> List[Node]:
        """
        returns a complete list of graph nodes, sorted in a topological order
        """
        g = nx.DiGraph()

        nodes = self.nodes
        for node in nodes:
            g.add_node(node)

        for node in nodes:
            for parent in node.get_parents():
                g.add_edge(parent, node)

        sorted_list = nx.topological_sort(g)

        return list(sorted_list)



from .BayesianEstimation import BayesianEstimation
from .CalibrateSystem import CalibrateSystem
from .ExternalScan import ExternalScan
from .FinalDataAnalysis import FinalDataAnalysis
from .QPUcircuitRunner import QPUcircuitRunner
from .TaskList import TaskList



"""
Generates python class for the workflow module based on module schema .json
"""

import sys
import os
import json
from jinja2 import Template
from .enum_types import InputType

if __name__ == "__main__":
    if len(sys.argv) <= 1:
        raise ValueError("Enter module names")

    template_filename = os.path.dirname(os.path.realpath(__file__))
    template_filename = os.path.join(template_filename, "node_template.py.jinja")
    with open(template_filename, "r") as file:
        node_template = Template(file.read())

    for node_name in sys.argv[1:]:
        full_path_in = os.path.join(os.path.join(os.getcwd(), "entropynodes"), "schema")
        full_path_in = os.path.join(full_path_in, f"{node_name}.json")

        with open(full_path_in, encoding="utf-8", mode="r") as f:
            schema = json.loads(f.read())

        full_path_out = os.path.join(
            os.path.join(os.getcwd(), "entropynodes"), "library"
        )
        if not os.path.exists(full_path_out):
            os.makedirs(full_path_out)

        # TODO: handle properly Inputs instances in Inputs lists, and multiple
        # Inputs instances (currently takes only first instance)

        input_type = {}
        for key, value in schema["inputs"][0]["type"].items():
            input_type[key] = InputType(value).name

        with open(
            os.path.join(full_path_out, f"{node_name}.py"), encoding="utf-8", mode="w"
        ) as f:
            f.write(
                node_template.render(
                    name=schema["name"],
                    description=schema["description"],
                    command=schema["command"],
                    bin=schema["bin"],
                    inputunits=schema["inputs"][0]["units"],
                    inputdescription=schema["inputs"][0]["description"],
                    inputtype=input_type,
                    outputdescription=schema["outputs"][0]["description"],
                    outputunits=schema["outputs"][0]["units"],
                    node_icon=schema["icon"],
                )
            )

        # load all the classes defined in library directory

        init_file = ""
        for py in [
            f[:-3]
            for f in os.listdir(full_path_out)
            if f.endswith(".py") and f != "__init__.py"
        ]:
            init_file += f"from .{py} import {py}\n"

        with open(os.path.join(full_path_out, "__init__.py"), "w") as f:
            f.write(init_file)

        with open(
            os.path.join(os.path.join(os.getcwd(), "entropynodes"), "__init__.py"), "w"
        ) as f:
            f.write("")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['entropylab',
 'entropylab.cli',
 'entropylab.components',
 'entropylab.dashboard',
 'entropylab.dashboard.pages',
 'entropylab.dashboard.pages.components',
 'entropylab.dashboard.pages.params',
 'entropylab.dashboard.pages.results',
 'entropylab.flame',
 'entropylab.flame.example.complex',
 'entropylab.flame.example.complex.entropynodes',
 'entropylab.flame.example.complex.entropynodes.library',
 'entropylab.flame.example.simple',
 'entropylab.flame.example.simple.entropynodes',
 'entropylab.flame.example.simple.entropynodes.library',
 'entropylab.flame.execute',
 'entropylab.flame.template.empty',
 'entropylab.flame.template.node',
 'entropylab.pipeline',
 'entropylab.pipeline.api',
 'entropylab.pipeline.results_backend',
 'entropylab.pipeline.results_backend.sqlalchemy',
 'entropylab.pipeline.results_backend.sqlalchemy.alembic',
 'entropylab.pipeline.results_backend.sqlalchemy.alembic.versions',
 'entropylab.quam']

package_data = \
{'': ['*'],
 'entropylab.dashboard': ['assets/*', 'assets/images/*'],
 'entropylab.flame.example.complex.entropynodes': ['schema/*'],
 'entropylab.flame.example.simple.entropynodes': ['schema/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'alembic>=1.6.5,<2.0.0',
 'asyncpg>=0.25,<0.26',
 'bokeh>=2.3.0,<3.0.0',
 'celery>=5.2.3,<6.0.0',
 'dash-bootstrap-components>=1.0.0,<2.0.0',
 'dash>=2.4.1,<3.0.0',
 'dill>=0.3.3,<0.4.0',
 'distro>=1.7.0,<2.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'fastapi>=0.75,<0.78',
 'graphviz>=0.16,<0.17',
 'h5py>=3.3.0,<4.0.0',
 'hiredis>=2.0.0,<3.0.0',
 'hupper>=1.10.3,<2.0.0',
 'jsonpickle>=2.0.0,<3.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'msgpack>=1.0.3,<2.0.0',
 'munch>=2.5.0,<3.0.0',
 'networkx>=2.6.0,<3.0.0',
 'numpy>=1.19,<2.0',
 'pandas>=1.2.3,<2.0.0',
 'param>=1.10.1,<2.0.0',
 'pika>=1.2.0,<2.0.0',
 'psutil>=5.9.0,<6.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pytz>=2021.3,<2023.0',
 'pyzmq>=22.3,<24.0',
 'qualang-tools>=0.9.0,<0.10.0',
 'redis>=4.1.0,<5.0.0',
 'requests>=2.27.1,<3.0.0',
 'sqlalchemy>=1.4.0,<2.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'tinydb>=4.5.2,<5.0.0',
 'tzlocal>=4.1,<5.0',
 'waitress>=2.1.2,<3.0.0']

entry_points = \
{'console_scripts': ['entropy = entropylab.cli.main:main',
                     'n3p = entropylab.cli.main:main']}

setup_kwargs = {
    'name': 'entropylab',
    'version': '0.11.1',
    'description': '',
    'long_description': '![PyPI](https://img.shields.io/pypi/v/entropylab)\n[![discord](https://img.shields.io/discord/806244683403100171?label=QUA&logo=Discord&style=plastic)](https://discord.gg/7FfhhpswbP)\n\n[![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.4785097.svg)](https://doi.org/10.5281/zenodo.4785097)\n\n# Entropy\n\nEntropy is a lab workflow management package built for, but not limited-to, streamlining the process of running quantum information processing experiments. \n\nCheck out our [docs](https://docs.entropy-lab.io) for more information\n\nEntropy is built to solve a few major hurdles in experiment design: \n\n1. Building, maintaining and executing complex experiments\n2. Data collection\n3. Device management\n4. Calibration automation\n\nTo tackle these problems, Entropy is built around the central concept of a graph structure. The nodes of a graph give us a convenient way \nto brake down experiments into stages and to automate some tasks required in each node. For example data collection is automated, at least in part, \nby saving node data and code to a persistent database. \n\nDevice management is the challenge of managing the state and control of a variety of different resources. These include, but are not limited to, lab instruments. \nThey can also be computational resources, software resources or others. Entropy is built with tools to save such resources to a shared database and give nodes access to \nthe resources needed during an experiment. \n\nPerforming automatic calibration is an important reason why we built Entropy. This could be though of as the use case most clearly benefiting from shared resources, persistent \nstorage of different pieced of information and the graph structure. If the final node in a graph is the target experiment, then all the nodes between the root and that node are often \ncalibration steps. The documentation section will show how this can be done. \n\nThe Entropy system is built with concrete implementations of the various parts (database backend, resource management and others) but is meant to be completely customizable. Any or every part of the system can be tailored by end users. \n\n## Modules \n\n- ***Pipeline*** : A simple execution engine for a collection of nodes. Allows passing data between nodes and saving results to a database. Also includes a dashboard for viewing results. \n- ***Flame*** : An actor model execution engine \n- ***QuAM*** : The Quantum Abstract Machine. An abstraction layer above QPU to simplify experiment authoring and parameter management.\n \n\n## Installation\n\nInstallation is done from pypi using the following command\n\n```shell\npip install entropylab\n```\n\n## Versioning and the Alpha release \n\nThe current release of Entropy is version 0.x.x. You can learn more about the Entropy versioning scheme in the versioning\ndocument. There will more than likely be breaking changes to the API for a while until we learn how things should be done.\n\n',
    'author': 'Tal Shani',
    'author_email': 'tal@quantum-machines.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/entropy-lab/entropy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



# ==================== DEFINE NODE ====================
import entropylab.flame.nodeio as nodeio
import time

nodeio.context(
    name="GrumpyAdministrator",
    description="divides task lisk and sends them to others",
    icon="bootstrap/person-lines-fill.svg",
)

inputs = nodeio.Inputs()
inputs.state(
    "customers", units="list of strings", description="all customers we have today"
)
inputs.state("clerk_salary", units="k$", description="workforce demands")

outputs = nodeio.Outputs()
outputs.define(
    "clerk_request",
    units="string",
    description="notifies connected clerk to do work",
    retention=0,
)

nodeio.register()

# ==================== DRY RUN DATA ====================

inputs.set(customers=["Alice", "Bob"])
inputs.set(clerk_salary=23.2)

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:
    today_work = inputs.get("customers")
    for i, customer in enumerate(today_work):
        outputs.set(clerk_request=customer)
        print(f"Here comes {customer} (number = {i})")
        # budget_requests = inputs.get("clerk_salary")
        # print(f"Clerk demands {budget_requests} k$")

    time.sleep(1)  # coffee break
    budget_requests = inputs.get("clerk_salary")
    print(f"Clerk demands {budget_requests} k$")

    nodeio.terminate_workflow()



import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from typing import Optional

import hupper
import waitress
from hupper.interfaces import ILogger

from entropylab.config import settings
from entropylab.dashboard.app import build_dashboard_app
from entropylab.logger import logger
from entropylab.pipeline.results_backend.sqlalchemy.project import dashboard_log_path


def serve_dashboard(
    path: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    debug: Optional[bool] = None,
):
    """Serves our "dashboard" app using waitress and opens it in a browser.

    :param path: The path to the Entropy project to connect to the dashboard.
    :param host: The host name from which to server the app. Defaults to "127.0.0.1".
    :param port: The port name from which to server the app. Defaults to 8050.
    :param debug: Start the dashboard in debug mode. Defaults to False.
    """

    # debug mode
    if debug:
        file_handler = _create_file_handler(dashboard_log_path(path))
        _set_to_debug(logger, file_handler)
        _set_to_debug(logging.getLogger("waitress"), file_handler)
        _set_to_debug(logging.getLogger("hupper"), file_handler)
        hupper_logger = HupperLogger(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
        logging.getLogger("waitress").setLevel(logging.INFO)
        hupper_logger = HupperLogger(logging.INFO)

    # defaults for args
    if host is None:
        host = settings.get("dashboard.host", "127.0.0.1")
    if port is None:
        port = settings.get("dashboard.port", 8050)
    if debug is None:
        debug = settings.get("dashboard.debug", False)

    # voodoo!
    sys.path.append(os.path.abspath(path))

    # building our Dash app
    # noinspection PyShadowingNames
    app = build_dashboard_app(path)
    app.enable_dev_tools(debug=True)

    # hot reloading using hupper
    worker_kwargs = dict(path=path, host=host, port=port, debug=debug)
    hupper.start_reloader(
        "entropylab.dashboard.serve_dashboard",
        worker_kwargs=worker_kwargs,
        reload_interval=0,
        logger=hupper_logger,
    )

    # finally
    waitress.serve(app.server, host=host, port=port)


def _create_file_handler(log_path):
    file_handler = RotatingFileHandler(
        log_path, maxBytes=10 * 1024 * 1024, backupCount=7
    )
    file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    return file_handler


# noinspection PyShadowingNames
def _set_to_debug(logger: logging.Logger, file_handler):
    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)


class HupperLogger(ILogger):
    def __init__(self, level):
        self.logger = logging.getLogger("hupper")
        logger.setLevel(level)

    def error(self, msg):
        self.logger.error(msg)

    def info(self, msg):
        self.logger.info(msg)

    def debug(self, msg):
        self.logger.debug(msg)



import dash_bootstrap_components as dbc
from dash import html, dcc, dash_table

from entropylab.dashboard.pages.components.footer import footer
from entropylab.dashboard.pages.components.top_bar import top_bar
from entropylab.dashboard.pages.params.utils import (
    param_store_to_commits_df,
    param_store_to_df,
)
from entropylab.dashboard.theme import (
    table_style_header,
    table_style_filter,
    table_style_data,
    table_style_cell,
    table_active_cell_conditional,
)
from entropylab.pipeline.api.param_store import ParamStore

REFRESH_INTERVAL_IN_MILLIS = 3000


def build_layout(path: str, param_store: ParamStore):

    return dbc.Container(
        fluid=True,
        className="main",
        children=[
            dcc.Interval(
                id="interval", interval=REFRESH_INTERVAL_IN_MILLIS, n_intervals=0
            ),
            top_bar(path),
            dbc.Row(
                children=[
                    dbc.Col(
                        [
                            dbc.Row(
                                [
                                    dbc.Col(html.H5("Params and their values")),
                                    dbc.Col(
                                        dbc.Collapse(
                                            dbc.Badge(
                                                "Dirty state",
                                                color="white",
                                                text_color="danger",
                                                id="dirty-badge",
                                                className="border ms-1",
                                                pill=True,
                                            ),
                                            is_open=False,
                                            id="dirty-badge-collapse",
                                        )
                                    ),
                                ]
                            ),
                            dbc.Row(
                                dbc.InputGroup(
                                    [
                                        dbc.Button(
                                            "➕ Add",
                                            color="primary",
                                            id="add-key-button",
                                            className="me-1",
                                        ),
                                        dbc.Button(
                                            "Commit",
                                            color="secondary",
                                            id="commit-button",
                                        ),
                                        dbc.Input(
                                            placeholder="Enter commit label",
                                            id="commit-label",
                                        ),
                                        dbc.Button(
                                            "Load temp",
                                            id="load-temp-button",
                                            className="me-1",
                                        ),
                                        dbc.Button(
                                            "Save temp",
                                            id="save-temp-button",
                                            className="me-1",
                                        ),
                                    ]
                                )
                            ),
                            dash_table.DataTable(
                                columns=[
                                    dict(
                                        name="key",
                                        id="key",
                                        type="text",
                                        editable=True,
                                    ),
                                    dict(
                                        name="value",
                                        id="value",
                                        type="any",
                                        editable=True,
                                    ),
                                    dict(
                                        name="tag",
                                        id="tag",
                                        type="text",
                                        editable=True,
                                    ),
                                ],
                                data=param_store_to_df(param_store).to_dict("records"),
                                id="data-table",
                                sort_action="native",
                                row_deletable=True,
                                filter_action="native",
                                editable=True,
                                fixed_rows={"headers": True, "data": 0},
                                style_data=table_style_data,
                                style_filter=table_style_filter,
                                style_header=table_style_header,
                                style_cell=table_style_cell,
                                style_data_conditional=[table_active_cell_conditional],
                            ),
                        ],
                        width=6,
                    ),
                    dbc.Col(
                        [
                            dbc.Row(html.H5("Commits")),
                            dbc.Row(
                                dbc.Col(
                                    dbc.Button("Checkout", id="checkout-button"),
                                    width="1",
                                ),
                            ),
                            dbc.Row(
                                dash_table.DataTable(
                                    columns=[
                                        dict(
                                            name="id",
                                            id="commit_id",
                                            type="text",
                                        ),
                                        dict(
                                            name="label",
                                            id="commit_label",
                                            type="text",
                                        ),
                                        dict(
                                            name="time",
                                            id="commit_time",
                                            type="datetime",
                                        ),
                                    ],
                                    data=param_store_to_commits_df(param_store).to_dict(
                                        "records"
                                    ),
                                    sort_action="native",
                                    sort_mode="multi",
                                    id="commit-table",
                                    filter_action="native",
                                    row_selectable="single",
                                    sort_by=[
                                        {
                                            "column_id": "commit_time",
                                            "direction": "desc",
                                        }
                                    ],
                                    fixed_rows={"headers": True, "data": 0},
                                    editable=True,
                                    style_data=table_style_data,
                                    style_filter=table_style_filter,
                                    style_header=table_style_header,
                                    style_cell=table_style_cell,
                                    style_data_conditional=[
                                        table_active_cell_conditional
                                    ],
                                )
                            ),
                        ],
                        width=6,
                    ),
                ],
            ),
            footer(),
        ],
    )



from datetime import datetime
from typing import List, Dict

import numpy as np
from plotly import express as px

from entropylab.pipeline.api.data_reader import FigureRecord
from entropylab.pipeline.api.errors import EntropyError


def auto_plot(experiment_id: int, data) -> FigureRecord:
    if isinstance(data, dict):
        return _auto_plot_from_dict(experiment_id, data)
    elif isinstance(data, list):
        plot = _auto_plot_from_list(data)
    elif isinstance(data, np.ndarray):
        plot = _auto_plot_from_ndarray(data)
    elif isinstance(data, int) or isinstance(data, float):
        plot = _auto_plot_from_list([data])
    else:
        raise EntropyError(
            "Only lists, dicts and ndarrays can be auto-plotted at this time"
        )
    plot.experiment_id = experiment_id
    plot.id = 0
    return plot


def _auto_plot_from_dict(experiment_id: int, data: Dict) -> FigureRecord:
    if len(data) > 0:
        first = list(data.values())[0]  # arbitrarily plot "first" value
        return auto_plot(experiment_id, first)
    else:
        raise EntropyError("Cannot auto-plot an empty dict")


def _auto_plot_from_list(data: List) -> FigureRecord:
    if not data:
        raise EntropyError("Cannot auto-plot None")
    if len(data) == 0:
        raise EntropyError("Cannot auto-plot an empty list")
    if _list_is_all_numeric(data):
        return _circle_from_list(data)
    elif _list_contains_one_list_of_scalars(data):
        return _circle_from_list(data[0])
    elif _list_contains_two_lists_of_equal_lengths(data):
        return _circle_from_xy(data[0], data[1])
    else:
        return _imshow_from_2d(data)


def _auto_plot_from_ndarray(data: np.ndarray) -> FigureRecord:
    array: np.ndarray = data
    if not array.any():
        raise EntropyError("Cannot auto-plot an empty ndarray")
    if not np.issubdtype(array.dtype, np.number):
        raise EntropyError("Cannot auto-plot a non-numeric ndarray")
    if _ndarray_is_1d(array):
        return _circle_from_list(array.tolist())
    elif _ndarray_contains_a_single_list_of_scalars(array):
        return _circle_from_list(array[0].tolist())
    elif _ndarray_contains_two_lists_of_equal_lengths(array):
        return _circle_from_xy(array[0], array[1])
    else:
        return _imshow_from_2d(data)


# List helper functions


def _is_numeric(first):
    return isinstance(first, int) or isinstance(first, float)


def _list_is_all_numeric(lst):
    return len(lst) > 0 and all(_is_numeric(x) for x in lst)


def _is_all_lists(lst):
    return all(isinstance(x, List) for x in lst)


def _list_contains_two_lists_of_equal_lengths(data):
    if not _is_all_lists(data):
        raise EntropyError("Auto-plot does not support mixed types")
    if len(data) > 2:
        return False
    if not len(data[0]) == len(data[1]):
        raise EntropyError("Cannot auto-plot 2 sub-lists of unequal lengths")
    if len(data[0]) == 0:
        raise EntropyError("Cannot auto-plot 2 empty sub-lists")
    return True


def _list_contains_one_list_of_scalars(data):
    return (
        len(data) == 1 and isinstance(data[0], list) and _list_is_all_numeric(data[0])
    )


# ndarray helper functions


def _ndarray_is_1d(array):
    return array.ndim == 1 and _list_is_all_numeric(array.tolist())


def _ndarray_contains_a_single_list_of_scalars(array):
    return array.shape[0] == 1 and _list_is_all_numeric(array[0].tolist())


def _ndarray_contains_two_lists_of_equal_lengths(array):
    return (
        array.ndim == 2
        and array.shape[0] == 2
        and _list_is_all_numeric(array[0].tolist())
        and _list_is_all_numeric(array[1].tolist())
    )


def _ndarray_is_2d(array):
    return array.ndim == 2


def _ndarray_is_3d_rgb(array):
    return array.ndim == 3 and array.shape[2] == 3  # Last dim is rgb triples


# helper functions for plotting


def _circle_from_xy(x, y):
    return FigureRecord(
        experiment_id=0, id=0, figure=px.scatter(x=x, y=y), time=datetime.now()
    )


def _xs_from(lst):
    return list(range(len(lst)))


def _circle_from_list(lst):
    x = _xs_from(lst)
    y = lst
    return _circle_from_xy(x, y)


def _imshow_from_2d(data):
    return FigureRecord(
        experiment_id=0, id=0, figure=px.imshow(data), time=datetime.now()
    )



import sys
import logging
from logging import StreamHandler, Formatter, INFO, WARN, ERROR

logger = logging.getLogger("entropy")
logger.setLevel(INFO)
_handler = StreamHandler(sys.stdout)
_handler.setFormatter(Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
logger.addHandler(_handler)

__all__ = ["INFO", "WARN", "ERROR", "logger"]



"""Initial migration

Revision ID: 1318a586f31d
Revises:
Create Date: 2021-08-01 13:45:43.059646+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
from sqlalchemy.engine.reflection import Inspector

revision = "1318a586f31d"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()
    # ### commands auto generated by Alembic - please adjust! ###
    if "Experiments" not in tables:
        op.create_table(
            "Experiments",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("label", sa.String(), nullable=True),
            sa.Column("script", sa.String(), nullable=True),
            sa.Column("start_time", sa.DATETIME(), nullable=False),
            sa.Column("end_time", sa.DATETIME(), nullable=True),
            sa.Column("user", sa.String(), nullable=True),
            sa.Column("story", sa.String(), nullable=True),
            sa.Column("success", sa.Boolean(), nullable=True),
            sa.PrimaryKeyConstraint("id"),
        )
    if "Resources" not in tables:
        op.create_table(
            "Resources",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("update_time", sa.DATETIME(), nullable=False),
            sa.Column("name", sa.String(), nullable=True),
            sa.Column("driver", sa.String(), nullable=True),
            sa.Column("module", sa.String(), nullable=True),
            sa.Column("class_name", sa.String(), nullable=True),
            sa.Column("version", sa.String(), nullable=True),
            sa.Column(
                "driver_type",
                sa.Enum("Packaged", "Local", name="drivertype"),
                nullable=True,
            ),
            sa.Column("args", sa.String(), nullable=True),
            sa.Column("kwargs", sa.String(), nullable=True),
            sa.Column("deleted", sa.Boolean(), nullable=True),
            sa.Column("number_of_experiment_args", sa.Integer(), nullable=True),
            sa.Column("keys_of_experiment_kwargs", sa.String(), nullable=True),
            sa.Column("cached_metadata", sa.String(), nullable=True),
            sa.PrimaryKeyConstraint("id"),
        )
    if "Debug" not in tables:
        op.create_table(
            "Debug",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("experiment_id", sa.Integer(), nullable=True),
            sa.Column("python_env", sa.String(), nullable=True),
            sa.Column("python_history", sa.String(), nullable=True),
            sa.Column("station_specs", sa.String(), nullable=True),
            sa.Column("extra", sa.String(), nullable=True),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )
    if "ExperimentMetadata" not in tables:
        op.create_table(
            "ExperimentMetadata",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("experiment_id", sa.Integer(), nullable=True),
            sa.Column("stage", sa.Integer(), nullable=True),
            sa.Column("label", sa.String(), nullable=True),
            sa.Column("time", sa.DATETIME(), nullable=False),
            sa.Column("data", sa.BLOB(), nullable=True),
            sa.Column(
                "data_type",
                sa.Enum("Pickled", "String", "Npy", name="resultdatatype"),
                nullable=True,
            ),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )
    if "Nodes" not in tables:
        op.create_table(
            "Nodes",
            sa.Column("experiment_id", sa.Integer(), nullable=False),
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("label", sa.String(), nullable=True),
            sa.Column("start", sa.DATETIME(), nullable=False),
            sa.Column("is_key_node", sa.Boolean(), nullable=True),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("experiment_id", "id"),
        )
    if "Plots" not in tables:
        op.create_table(
            "Plots",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("experiment_id", sa.Integer(), nullable=True),
            sa.Column("plot_data", sa.BLOB(), nullable=True),
            sa.Column(
                "data_type",
                sa.Enum("Pickled", "String", "Npy", name="resultdatatype"),
                nullable=True,
            ),
            sa.Column("generator_module", sa.String(), nullable=True),
            sa.Column("generator_class", sa.String(), nullable=True),
            sa.Column("time", sa.DATETIME(), nullable=True),
            sa.Column("label", sa.String(), nullable=True),
            sa.Column("story", sa.String(), nullable=True),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )
    if "ResourcesSnapshots" not in tables:
        op.create_table(
            "ResourcesSnapshots",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("update_time", sa.DATETIME(), nullable=False),
            sa.Column("name", sa.String(), nullable=True),
            sa.Column("driver_id", sa.Integer(), nullable=True),
            sa.Column("state", sa.String(), nullable=True),
            sa.ForeignKeyConstraint(
                ["driver_id"], ["Resources.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )
    if "Results" not in tables:
        op.create_table(
            "Results",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("experiment_id", sa.Integer(), nullable=True),
            sa.Column("stage", sa.Integer(), nullable=True),
            sa.Column("story", sa.String(), nullable=True),
            sa.Column("label", sa.String(), nullable=True),
            sa.Column("time", sa.DATETIME(), nullable=False),
            sa.Column("data", sa.BLOB(), nullable=True),
            sa.Column(
                "data_type",
                sa.Enum("Pickled", "String", "Npy", name="resultdatatype"),
                nullable=True,
            ),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("Results")
    op.drop_table("ResourcesSnapshots")
    op.drop_table("Plots")
    op.drop_table("Nodes")
    op.drop_table("ExperimentMetadata")
    op.drop_table("Debug")
    op.drop_table("Resources")
    op.drop_table("Experiments")
    # ### end Alembic commands ###



from abc import abstractmethod
from typing import Callable, Dict
from munch import Munch

from entropylab.pipeline.api.in_process_param_store import InProcessParamStore
from qualang_tools.config import ConfigBuilder
from qualang_tools.config.parameters import ConfigVars
from qm.QuantumMachinesManager import QuantumMachinesManager


class QuAMCore(object):
    def __init__(self, path: str):
        """Core QUAM object
        :param path: Path of the entropy DB
        :type path: str
        """
        self.path = path
        self.param_store = InProcessParamStore(path)


class QuAMManager(QuAMCore):
    def __init__(self, path: str, host: str = "127.0.0.1"):
        """Admin
        :param path: Path of the entropy DB
        :type path: str
        :param host: QM manager url
        :type host: str
        """
        super().__init__(path=path)
        self.host = host
        self._config_builder = ConfigBuilder()
        self._config_vars = ConfigVars()

    def generate_config(self):
        """Returns the QUA configuration"""
        self._config_builder = ConfigBuilder()
        self.prepare_config(self._config_builder)
        self._set_config_vars()
        return self._config_builder.build()

    @abstractmethod
    def prepare_config(self, cb: ConfigBuilder) -> None:
        """method to prepare the ConfigBuilder, needs to be implemented by the QuAMManager"""
        pass

    def parameter(self, var: str, setter: Callable = None):
        """Returns a parameter with the given name and setter, all parameters are saved
        in the ParamStore
        :return: a parameter
        :rtype: Parameter
        """
        if var not in self.param_store.keys():
            self.param_store[var] = None
        return self._config_vars.parameter(var, setter=setter)

    def _set_config_vars(self):
        """Sets the parameter values according to the key, value pairs in the param store"""
        _dict = {}
        for k in self.param_store.keys():
            if not callable(self._config_vars.parameter(k)._value):
                _dict[k] = self.param_store[k]
        self._config_vars.set(**_dict)

    def open_quam(self):
        """Returns an instance of QuAM
        :return: a quam
        :rtype: QuAM
        """
        self._set_config_vars()
        return QuAM(path=self.path, config=self.generate_config())


class QuAM(QuAMCore):
    def __init__(self, path: str, config: dict = None, host: str = "127.0.0.1"):
        """User class to facilitate writing and execution of QUA programs
        :param path: path to the entropy DB
        :type path: str
        :param config: a QUA configuration, defaults to an empty dictionary
        :type config: Dict, optional
        :param host: host url for Quantum Machine Manager, defaults to local host
        :type host: str, optional
        """
        super().__init__(path=path)
        self.config = config if config is not None else {}
        self.host = host
        self.qmm = QuantumMachinesManager(host)

    @property
    def elements(self):
        """Returns a Munch instance of element keys available in the QUA configuration"""
        return dict_keys_to_munch(self.config["elements"])

    @property
    def pulses(self):
        """Returns a Munch instance of pulses available in the QUA configuration"""
        return dict_keys_to_munch(self.config["pulses"])

    @property
    def integration_weights(self):
        """Returns a Munch instance of integration weights available in the QUA configuration"""
        return dict_keys_to_munch(self.config["integration_weights"])


def dict_keys_to_munch(d: Dict):
    elms = Munch()
    for k in d.keys():
        elms[k] = k
    return elms



"""Add col saved_in_hdf5

Revision ID: 04ae19b32c08
Revises: 1318a586f31d
Create Date: 2021-08-01 14:18:43.052881+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
from sqlalchemy import text

revision = "04ae19b32c08"
down_revision = "1318a586f31d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "Results",
        sa.Column(
            "saved_in_hdf5",
            sa.Boolean(),
            nullable=False,
            default=False,
            server_default=text("False"),
        ),
    )
    op.add_column(
        "ExperimentMetadata",
        sa.Column(
            "saved_in_hdf5",
            sa.Boolean(),
            nullable=False,
            default=False,
            server_default=text("False"),
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("Results", "saved_in_hdf5")
    op.drop_column("ExperimentMetadata", "saved_in_hdf5")
    # ### end Alembic commands ###



import entropylab.flame.nodeio as nodeio
import time

# ==================== DEFINE NODE ====================

nodeio.context(
    name="CalibrateSystem",
    description="Makes sure that system is calibrated",
    icon="bootstrap/gear-fill.svg",
)

input = nodeio.Inputs()
input.state(
    "instrumentAddress",
    units="IP address",
    description="instrument IP address for calibration",
)

output = nodeio.Outputs()
output.define(
    "calibration_status",
    units="calibrated, uncalibrated",
    description="Calibration status",
    retention=0,
)

nodeio.register()  # enables this to be used as part of workflow

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:
    a = False
    print("calibrating")

    time.sleep(0.5)
    new_state = True

    if new_state:
        if new_state != a:
            output.set(calibration_status="calibrated")
    else:
        if new_state != a:
            output.set(calibration_status="uncalibrated")
    a = new_state

# ================= =================== =================



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["CheerfulNode"]


class CheerfulNode(object):
    def __init__(self, workflow_node_unique_name, customer=None, weather=None):
        """greets customers

        :param customer: (human - STREAM) one person at a time
        :param weather: (best guess - STATE) How is weather today
        """
        self._command = "python3"
        self._bin = "cheerful_node.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/person-circle.svg"
        self._inputs = _Inputs(
            customer=customer,
            weather=weather,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, customer=None, weather=None):
        self._inputs = Inputs()

        self._inputs.state(
            "customer", description="one person at a time", units="human"
        )
        self._inputs.set(customer=customer)

        self._inputs.state(
            "weather", description="How is weather today", units="best guess"
        )
        self._inputs.set(weather=weather)

    @property
    def customer(self):
        """Input: one person at a time (human)"""
        return self._inputs.get("customer")

    @customer.setter
    def customer(self, value):
        """Input: one person at a time (human)"""
        self._inputs.set(customer=value)

    @property
    def weather(self):
        """Input: How is weather today (best guess)"""
        return self._inputs.get("weather")

    @weather.setter
    def weather(self, value):
        """Input: How is weather today (best guess)"""
        self._inputs.set(weather=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "requested_salary",
        ]

    @property
    def requested_salary(self):
        """Output: requested fees from administrator
        :return: (k$)
        """
        return "#" + self._name + "/requested_salary"



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["GrumpyAdministrator"]


class GrumpyAdministrator(object):
    def __init__(self, workflow_node_unique_name, customers=None, clerk_salary=None):
        """divides task lisk and sends them to others

        :param customers: (list of strings - STATE) all customers we have today
        :param clerk_salary: (k$ - STATE) workforce demands
        """
        self._command = "python3"
        self._bin = "grumpy_administrator.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/person-lines-fill.svg"
        self._inputs = _Inputs(
            customers=customers,
            clerk_salary=clerk_salary,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, customers=None, clerk_salary=None):
        self._inputs = Inputs()

        self._inputs.state(
            "customers",
            description="all customers we have today",
            units="list of strings",
        )
        self._inputs.set(customers=customers)

        self._inputs.state("clerk_salary", description="workforce demands", units="k$")
        self._inputs.set(clerk_salary=clerk_salary)

    @property
    def customers(self):
        """Input: all customers we have today (list of strings)"""
        return self._inputs.get("customers")

    @customers.setter
    def customers(self, value):
        """Input: all customers we have today (list of strings)"""
        self._inputs.set(customers=value)

    @property
    def clerk_salary(self):
        """Input: workforce demands (k$)"""
        return self._inputs.get("clerk_salary")

    @clerk_salary.setter
    def clerk_salary(self, value):
        """Input: workforce demands (k$)"""
        self._inputs.set(clerk_salary=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "clerk_request",
        ]

    @property
    def clerk_request(self):
        """Output: notifies connected clerk to do work
        :return: (string)
        """
        return "#" + self._name + "/clerk_request"
